
variable versionNumber "v05.30"
################################################################################
# CxDiag.tcl
# ------------------------------------------------------------------------------
# This script is provided as a utility supporting Paragon-neo and Paragon-100G.
# It is NOT intended as example code for controlling these instruments.
# No part of this script should be used independently of this script without
# the explicit written permission of Calnex Solutions Ltd.
################################################################################
# Version 05.30
#   Fixed issue with Rj45_Lowrate being erroneously detected as an optical interface.
#   Stopped attempted download of neo/Logfiles/MemDiagnostics folder. In future this  
#   should be reinstated properly.
# Version 05.20
#   Added support for instruments that do not have a PTP option fitted
# Version 05.10
#   Now downloads embedded logs
# Version 05.00
#   Added ability to download reports
# Version 04.02
#	Removed dump of slave settings - under some circumstance this causes a
#	crash. Just removed for the moment - put it back at some point later.
# Version 04.01
#	Added support for additional pages for QSFP optics.
# Version 04.00
#	Re-factored register read/write functionality
#	All versions are now combined into one supporting all functionality that
#	was previously delivered as Basic/Full/Advanced
# Version 03.01
#	Added support for more interface types. This is dependent on software version.
#		The software auto-detects whether interfaces are supported.
#		For SFP1G, this is only supported on Paragon-neo v01.00 or later
# Version 03.00
#	Added more debug info to diagnostic file
#	Added FEC counters to the list of items that can be monitored
# Version 02.01
#	Added support for writing to optical module registers (for Calnex use only)
# Version 02.00
#	Added handling of CFP2 interfaces (to avoid crash)
#	Now supports software version that don't support de-coupling ports
#	Added logging
#	Added support for basic version that supports getting session files and
#		debug info only.
# 	Added support for cdf->pcap conversion (in Sessions tab only)
# Version 01.00
#	Initial release
################################################################################
source CxLogging.tcl
# Console logging disabled for release - seems to cause big delays sometimes
cxLogging TRUE FALSE

source CxInstrumentClass.tcl

# There must be a better way...
global paragon
global lastUsedIpAddressListFilename
# Mode is BASIC, FULL, ADVANCED
global cxDiagMode
#global btnGetSessionList
global fSessions
global fReports
global fDebugInfo
global fOpticalModuleInfo
global fIp
global fTabs
global fAllPacketCapture
global fTop
global fOpticalModuleEepromLabelFrame
global fPortMonitor
global fPortMonitorItemSelectionLabelFrame
global lblStatus
global downloadedDataFolderPrefix
global powerMeasurementRunning
global powerMeasurementStop
global portMonitorStop
global portMonitorRxPowerSelected
global portMonitorRxPowerAlarmsSelected
global portMonitorTxPowerSelected
global portMonitorLosSelected
global portMonitorBipsSelected
global sessionsDownloadGeneratePcapSelected

set downloadedDataFolderPrefix "CxDiagData_"
set powerMeasurementRunning FALSE
set powerMeasurementStop TRUE
set portMonitorStop TRUE

oo::class create CxPcapConvertor {
	variable RootFolder
	variable SessionFolder
	variable ConversionExe
	
	constructor {rootFolder sessionFolder} {
		cxLogDebug "[info level 0]"
		set RootFolder $rootFolder
		set SessionFolder $sessionFolder
		set ConversionExe [file join "." "pcapConvertor" "CdfToPcap.exe"]
	}
	
	destructor {}
	
	# Convert the PTP capture file in the session folder to pcap
	method convert {} {
		cxLogDebug "[info level 0]"
		# Check that the session contains a PTP capture
		set captureFile [file join "." "$SessionFolder" "Ptp1588100.CDF"]
		if {![file exists $captureFile]} {
			cxLogDebug "$captureFile does not exist"
			return "no PTP capture in Session"
		}
		# Convert the file
		set pcapFile [file join "." "$SessionFolder" "Ptp1588100.pcap"]
		cxLogDebug "Converting $captureFile to $pcapFile..."
		exec $ConversionExe $captureFile $pcapFile
		cxLogDebug "Converting $captureFile to $pcapFile... done"
		return "done"
	}
}

oo::class create CxDebugData {
	variable paragon
	variable date
	variable root
	variable sessions
	variable logFiles
	variable logEmbedded
	variable logControl
	variable noGpio
	
    constructor { instrument destFolderPrefix cxDiagMode} {
		cxLogDebug "[info level 0]"
		set paragon $instrument
		set sn [$paragon getSerialNumber]
		#puts $destFolderPrefix
		set folderName [file join "[pwd]" "$destFolderPrefix$sn"]

		#set root "[pwd]/$destFolderPrefix[$paragon getSerialNumber]"
		# HACK HACK HACK
		# There should be some way to identify subfolders and deal with these but as it
		set root $folderName
		set sessions "$root/Sessions"
		set logFiles "$root/Logs/LogFiles"
		set logEmbedded "$root/Logs/EmbeddedLogs"
		set logControl "$root/Logs/Control"
		file mkdir $sessions
		file mkdir $logFiles
		file mkdir $logEmbedded
		file mkdir $logControl
		
		set date [clock format [clock seconds] -format %Y-%m-%d]
		if {$cxDiagMode == "BASIC"} {
			set noGpio TRUE
		} else {
			set noGpio FALSE			
		}
	}
    destructor {}

	# Copied from https://wiki.tcl-lang.org/page/pdict%3A+Pretty+print+a+dict
	# Needed some syntax errors fixing though
	# Private function
	method Pdict {fid args} {
		#puts "Fid and args: $fid $args"
		set cmd pdict
		set usage "usage: $cmd ?maxlevel? dictionaryValue ?globPattern?..."
		if {[string is integer [lindex $args 0]]} {
			set maxlvl [lindex $args 0]
			set args [lrange $args 1 end]
		} else {
			set maxlvl [llength $args]
		}
		#puts "max lvl: $maxlvl"
		#puts "fid: $fid"

		foreach {dvar pat} $args break
		if {$dvar == ""} {error $usage}
		if {$pat  == ""} {set pat "*"}
		set args [lrange $args 2 end]
		upvar __pdict__level __pdict__level
		incr __pdict__level +1
		set sp [string repeat "  " [expr $__pdict__level-1]]
		if { [catch {dict keys $dvar $pat} keys] } {
			error "$cmd error: 'dictionaryValue' is no 'dict'\n → $usage\n  → $keys"
		} elseif {[llength $keys]} {
			set size [::tcl::mathfunc::max {*}[lmap k $keys {string length $k}]]
			foreach key $keys {
				set dsubvar [dict get $dvar $key]
				puts -nonewline $fid [format {%s%-*s} $sp $size $key]
				if {$__pdict__level < $maxlvl} {
					set isVal   [catch {dict keys $dsubvar} keys]
					if {[llength $keys] == 0} {
						puts $fid " = \{\}"
					} elseif {$isVal} {
						puts $fid " = $dsubvar"
					} else {
						puts $fid " \{"
						#puts "Calling pdict again"
						my Pdict $fid $maxlvl $dsubvar {*}$args
						puts $fid "$sp\}"
					}
				} else {
					puts $fid " = $dsubvar"
				}
			}
		} else {
		}
		if {$__pdict__level == 1} {
			unset -nocomplain __pdict__level
		} else {
			incr __pdict__level -1
		}
	}
	
	method WriteGpioStates {fid port ethernetSettings} {
		cxLogDebug "[info level 0]"
		set ifs [dict get $ethernetSettings "Interfaces"]
		if {[dict exists $ifs "Rj45_1G"]} {
			set ifs [dict remove $ifs "Rj45_1G"]
		}
		
		if {[dict exists $ifs "Rj45_LowRate"]} {
			set ifs [dict remove $ifs "Rj45_LowRate"]
		}
		
		set test "123"
		
		
		set interfaces [dict keys $ifs]
		# For testing
		#set interfaces "CFP2 $interfaces"

		foreach interface $interfaces {
			CxInstrumentOpticalModule create $interface $port $interface
			set gpio [$interface readGpio]
			set inserted "Not inserted"
			if {$gpio == 0} {
				set inserted "Inserted"				
			}
			if {![dict exists $ifs $interface]} {
				puts $fid "[format %-12s $interface:] $inserted (Fitted=NotSupported)"				
			} else {
				set fitted [dict get $ifs $interface "Fitted"]
				puts $fid "[format %-12s $interface:] $inserted (Fitted=$fitted)"
			}
			$interface destroy
		}
	}
	
	method writeDebugInfoFile {} {
		cxLogDebug "[info level 0]"

		set f "$root/debugInfo_$date.txt"
		set fid [open $f "w"]
	
		statusUpdate "Writing debug info file..." green
		puts $fid "Date\t\t\t = $date"
		puts $fid "Serial Number\t = [$paragon getSerialNumber]"
		puts $fid "Instrument Model\t = [$paragon getModel]"
		puts $fid "SW Version\t\t = [$paragon getSoftwareVersion]"
		puts $fid "OS Version\t\t = [$paragon getOsVersion]"
		puts $fid "Capabilities\t = [$paragon getCapabilities]"
		puts $fid "Port Coupling\t = [$paragon getPortCouplingState]"
		puts $fid "System Time\t\t = [$paragon getSystemTime]"
		puts $fid ""
		puts $fid "PORT 1 SETTINGS"
		set p1EthernetSettings [$paragon port1 getEthernetSettings]
		my Pdict $fid 10 $p1EthernetSettings
		puts $fid ""
		puts $fid "Port 1 GPIO"
		if {$noGpio} {
			puts $fid "GPIO not available in basic mode"			
		} else {
			my WriteGpioStates $fid "Port 1" $p1EthernetSettings
		}
		puts $fid ""
		puts $fid "PORT 2 SETTINGS"
		set p2EthernetSettings [$paragon port2 getEthernetSettings]
		my Pdict $fid 10 $p2EthernetSettings
		
		puts $fid ""
		puts $fid "Port 2 GPIO"
		if {$noGpio} {
			puts $fid "GPIO not available in basic mode"			
		} else {
			my WriteGpioStates $fid "Port 2" $p2EthernetSettings
		}
		puts $fid ""
		puts $fid "INSTRUMENT OPTIONS"
		cxLogDebug "INSTRUMENT OPTIONS"
		#my Pdict $fid 10 [$paragon getOptions]
		set options [$paragon getOptions]
		foreach opt $options {
			puts $fid "[format %-24s [lindex $opt 1]]\t[format %-64s [lindex $opt 3]]\t[format %-10s [lindex $opt 5]]"
		}

		puts $fid ""
		puts $fid "REFERENCES"
		cxLogDebug "REFERENCES"
		set refs [$paragon getFrequencyReferenceSettings]
		my Pdict $fid 10 $refs
		set refs [$paragon getOnePpsReferenceSettings]
		my Pdict $fid 10 $refs
		

		puts $fid ""
		puts $fid "APP STATUS"
		cxLogDebug "APP STATUS"
		set appStatus [$paragon getAppStatus]
		foreach app $appStatus {
			puts $fid "---------------------------------------------------------"
			my Pdict $fid 10 $app
		}
		puts $fid ""
		puts $fid "APP STATUS - G.8273.2 (Generation)"
		my Pdict $fid 10 [$paragon getAppStatusConformanceGeneration]
		puts $fid "APP STATUS - G.8273.2 (Measurement)"
		my Pdict $fid 10 [$paragon getAppStatusConformanceMeasurement]
		puts $fid "APP STATUS - G.8273.2 (Stimulus)"
		my Pdict $fid 10 [$paragon getAppStatusConformanceStimulus]

		puts $fid ""
		puts $fid "MSE SETTINGS - GENERAL"
        if [$paragon isPtpOptionFitted] {
            set mseSettings {dutethernetcabledelay measurement/elapsedtime \
            				 oneppsrefcabledelay ptpprofile testmode timingperiod}
            foreach setting $mseSettings {
            	puts $fid "[format %-24s $setting:]\t [$paragon getMseSetting $setting]"
            }
        } else {
            puts $fid "\tPTP option is not fitted"
        }
		
		puts $fid ""
		puts $fid "MSE SETTINGS - MASTER"
		set masterSettings {altmaster clockaccuracy clockclass clockid clockmode \
							connectedslaves domain encapsulation frequencytraceable \
							leap59 leap61 logvariance multicast \
							multicast/announcemessage multicast/announcemessagerate \
							multicast/delayresponse multicast/ipaddress \
							multicast/macaddress \
							multicast/syncmessage multicast/syncmessagerate \
							portnumber priority1 priority2 \
							ptptimescale sourceipaddress sourcemacaddress \
							stepsremoved syncandfollowupcorrectionfield \
							timesource timetraceable unicast \
							unicast/maxannouncerate unicast/maxdelresprate \
							unicast/maxsyncrate unicast/slaves/announceenabled \
							unicast/slaves/syncenabled \
							unicast/useparagonslaveaddressforfirstslave \
							unicast/usesignaling utcoffset utcoffsetvalid \
							vlantaggingmode vlantags}
        if [$paragon isPtpOptionFitted] {
    		foreach setting $masterSettings {
    			puts $fid "[format %-24s $setting:]\t [$paragon getMseMasterSetting Master1 $setting]"
    		}
        } else {
            puts $fid "\tPTP option is not fitted"
        }

		puts $fid ""
		puts $fid "MSE SETTINGS - MASTER SLAVES"
		cxLogDebug "MSE SETTINGS - MASTER SLAVES"
        if [$paragon isPtpOptionFitted] {
    		set slaves [$paragon getMseMasterUnicastSlaves Master1]
    		foreach slave $slaves {
    			puts $fid "---------------------------------------------------------"
    			my Pdict $fid 10 $slave
    		}
        } else {
            puts $fid "\tPTP option is not fitted"
        }

		puts $fid ""
		puts $fid "MSE SETTINGS - SLAVES"
		cxLogDebug "MSE SETTINGS - SLAVES"
		set slaveSettings {clockid connectedmaster delayrequestcorrection domain \
							domain/coupled encapsulation encapsulation/coupled \
							messagingmode \
							multicast/delreqrate multicast/ipaddress \
							multicast/macaddress \
							multicast/masteripaddress multicast/mastermacaddress \
							multicast/mastermode \
							portnumber sourceipaddress sourcemacaddress \
							unicast/announceduration unicast/announcerate unicast/announcerenew \
							unicast/delrespduration unicast/delresprate unicast/delresprenew \
							unicast/masteripv4address unicast/mastermacaddress \
							unicast/masterstatus unicast/renewgrants unicast/requestperiod \
							unicast/syncduration unicast/syncrate unicast/syncrenew \
							unicast/usecalnexmasteraddress unicast/usesignaling \
							vlantaggingmode vlantags vlantagscoupled}

        if [$paragon isPtpOptionFitted] {
    		set slaves [$paragon getMseMasterUnicastSlaves Master1]
    		set slaveNum 1
    		foreach slave $slaves {
    			puts $fid "Slave: $slaveNum -------------------------------------------------"
    			foreach setting $slaveSettings {
    				# HACK Commented out - seems to cause a crash under some circumstances - need to track this down at some point
    				#puts $fid "[format %-24s $setting:]\t [$paragon getMseSlaveSetting $slaveNum $setting]"
    			}
    			set slaveNum [expr $slaveNum + 1]
    		}
        } else {
            puts $fid "\tPTP option is not fitted"
        }

		puts $fid ""
		puts $fid "CAT STATUS"
		cxLogDebug "CAT STATUS"
		puts $fid "---------------------------------------------------------"
		set catStatus [$paragon getCatStatus]
		my Pdict $fid 10 $catStatus
		#puts $fid "---------------------------------------------------------"
		#set catMeas [$paragon getCatMeasurements]
		#my Pdict $fid 10 $catMeas
			
		close $fid
		statusUpdate "Writing debug info file... done" black

	}
	method downloadData {instrumentFolder} {
		global sessionsDownloadGeneratePcapSelected
		
		cxLogDebug "[info level 0]"

		#puts "Session Folder: $instrumentFolder"	
		set sessionFolder "$sessions/$instrumentFolder"
		file mkdir $sessionFolder
		puts "Copying Session files..."
		statusUpdate "Downloading session files... this may take some time" green
		$paragon downloadSessionFiles $instrumentFolder $sessionFolder
		puts "Copying Log files..."
		statusUpdate "Downloading log files... this may take some time" green
		$paragon downloadLogs $logFiles $logEmbedded $logControl
		statusUpdate "Download complete" black
	
		cxLogDebug "GeneratePcap: $sessionsDownloadGeneratePcapSelected"
		if {$sessionsDownloadGeneratePcapSelected} {
			statusUpdate "Generating pcap..." green
			CxPcapConvertor create pcapConv $root $sessionFolder
			set ret [pcapConv convert]
			pcapConv destroy
			statusUpdate "Generating pcap... $ret" black
		}
	}
	method downloadLogs {} {
		cxLogDebug "[info level 0]"
		#file mkdir $sessionFolder
		#puts "Copying Session files..."
		#$paragon downloadSessionFiles $instrumentFolder $sessionFolder
		statusUpdate "Downloading log files... this may take some time" green
		$paragon downloadLogs $logFiles $logEmbedded $logControl
		statusUpdate "Downloading log files... done" black
	
	}
	method writeOpticalModuleInfo {opticalModule lblStatus} {
		#variable fOpticalModuleInfo
		cxLogDebug "[info level 0]"
		
		set portName [$opticalModule getPortName]
		set type [$opticalModule getType]
		set pages [$opticalModule getEepromPages]
		
		set f "$root/opticalModuleInfo_${portName}_${type}_$date.csv"
		set fid [open $f "w"]

		puts $fid "Serial Number, Instrument Model, SW Version, Transceiver Type, Port"
		puts $fid "[$paragon getSerialNumber], [$paragon getModel], [$paragon getSoftwareVersion], $type, $portName"
		puts $fid ""
		
		foreach page $pages {
			#puts $fid "Page, Addr, Value, Name"
			set regRange [$opticalModule getRegisterRange $page]
			set first [dict get $regRange first]
			set last [dict get $regRange last]
			set readOneAtATime False
			if {$readOneAtATime} {
    			puts $fid "Page, Addr, Value, Name"
				for {set addr $first} {$addr <= $last} {incr addr} {
					$lblStatus configure -text "Reading $type registers... page $page address $addr ($first to $last)"
					update
					statusUpdate "Reading $type registers... page $page address $addr ($first to $last)" green

					set regVals [$opticalModule readRegister $page $addr]
					set regVal [format "0x%02x" [lindex $regVals 0]]
					set regName [lindex $regVals 1]
					puts $fid "$page, $addr, $regVal, $regName"
				}
			} else {
    			puts $fid "Page, Addr, AddrHex, Value, ValueHex, Name"
				cxLogDebug "Using register stream reader"
				# For QSFP* devices, we don't know whether any given page
				# (apart from 00h) is supported. In addition, the stream reader
				# doesn't do any checking for valid pages. So, before we start
				# reading, make sure that the page is supported by the device.
				if {[$opticalModule isPageSupported $page]} {
					$opticalModule openRegisterReader $page $first $last
					for {set addr $first} {$addr <= $last} {incr addr} {
						$lblStatus configure -text "Reading $type registers... page $page address $addr ($first to $last)"
						update
						statusUpdate "Reading $type registers... page $page address $addr ($first to $last)" green
	
						set regVals [$opticalModule getRegisterReaderLine]
						set regAddrHex [format "0x%02x" [lindex $regVals 0]]
                        set regValDec [lindex $regVals 1]
                        set regValHex [format "0x%02X" $regValDec]
						set regName [lindex $regVals 2]
						puts $fid "$page, $addr, $regAddrHex, $regValDec, $regValHex, $regName"
					}
				} else {
					puts $fid "$page, Not Supported"
				}
			}
			puts $fid ""
		}
		close $fid	
		statusUpdate "Reading $type registers... address $addr ($first to $last)" black
	}
	
	method openPowerMeasurementFile {portName interface} {
		cxLogDebug "[info level 0]"
		set f "$root/opticalModulePower_${portName}_${interface}_$date.csv"
		set fid [open $f "w"]
		return $fid
	}
	method openPortMonitorFile {} {
		cxLogDebug "[info level 0]"
		set f "$root/portMonitor_$date.csv"
		set fid [open $f "w"]
		return $fid
	}
}

proc openFileExplorer {} {
	variable paragon
	variable downloadedDataFolderPrefix

	cxLogDebug "[info level 0]"
	
	set sn [paragon getSerialNumber]
	eval exec [auto_execok start] [list "$downloadedDataFolderPrefix$sn"] 
}

proc statusUpdate {status colour} {
	variable lblStatus
	$lblStatus configure -text $status -fg $colour
	cxLogDebug "[info level 0]"
	update
}
proc getLastSessions {sessions numSessions} {
	#set t [time {cxLogDebug "[info level 0]"} 1]
	#statusUpdate "getLastSessions: time to get last 20 - $t" black
	set n [expr $numSessions-1]
#	return [lrange $sessions end-$n end]
	set t [time {set ret [lrange $sessions 0 $n]} 1]
	cxLogDebug "getLastSessions: time to get last 20 - $t"
	return $ret
}
proc getLastReports {reports numReports} {
	#set t [time {cxLogDebug "[info level 0]"} 1]
	#statusUpdate "getLastSessions: time to get last 20 - $t" black
	set n [expr $numReports-1]
#	return [lrange $sessions end-$n end]
	set t [time {set ret [lrange $reports 0 $n]} 1]
	cxLogDebug "getLastReports: time to get last 20 - $t"
	return $ret
}

proc connect {snLabel ipCombobox} {
	variable paragon
	variable fSessions
	variable fReports
	variable fDebugInfo
	variable fPortMonitor
	variable fOpticalModuleInfo
	variable fTop
	variable fIp
	variable fTabs
	variable downloadedDataFolderPrefix
	variable lblStatus
	
	cxLogDebug "[info level 0]"

	statusUpdate "" "black"
	sessionsClear
	reportsClear
	debugInfoClear
	opticalModuleInfoClear
	allPacketCaptureClear
	$fIp.btnConnect configure -state disabled
	$fTabs state disabled
	$snLabel configure -text "" -fg black
	$fSessions.btnGetSessionList configure -state disabled
	$fReports.btnGetReportList configure -state disabled
	$fDebugInfo.btnGetDebugInfo configure -state disabled
	$fOpticalModuleInfo.btnGetOpticalModuleInterfaces configure -state disabled
	$fTop.btnOpenFileExplorer configure -state disabled

	set ipAddress [$ipCombobox get]
	#puts $ipAddress
	set ipAddr [string trim $ipAddress]
	statusUpdate "Connecting to $ipAddr" "black"
	# If p already exists (e.g. when user is looking at different instruments)
	# then we need to destroy the current instance and re-create
	if {[info commands paragon] != ""} {
		#puts [info commands p]
		paragon destroy
	}
	set err [catch {CxInstrument create paragon $ipAddr} res]
	#CxInstrument create paragon $ipAddr
	#puts $err
	if {$err != 0} {
		#ERROR
		cxLogDebug "Failed to connect: $res"
		tooltip::tooltip $ipCombobox "Can't connect. Please check your IP Address"
		statusUpdate "Connecting to $ipAddr... Failed. Please check your IP address." "red"

		$ipCombobox configure -foreground red
		set connStatus "Please check your IP Address"
		set connStatusFg "red"
		$snLabel configure -text $connStatus -fg $connStatusFg
		$fTop.lblFolderInfo configure -text ""
		$fIp.btnConnect configure -state normal
		$fTabs state !disabled
	} else {
		statusUpdate "Connecting to $ipAddr... Now connected." "green"
		addIpAddressToLastUsedList $ipAddress
		$ipCombobox configure -foreground black
		tooltip::tooltip $ipCombobox ""
		set sn [paragon getSerialNumber]
		$snLabel configure -text "[paragon getModel] $sn"
		$fSessions.btnGetSessionList configure -state normal
		$fReports.btnGetReportList configure -state normal
		$fDebugInfo.btnGetDebugInfo configure -state normal
		$fOpticalModuleInfo.btnGetOpticalModuleInterfaces configure -state normal
		$fTop.lblFolderInfo configure -text "Downloaded data will be stored in ./$downloadedDataFolderPrefix$sn"
		$fTop.btnOpenFileExplorer configure -state normal
		$fIp.btnConnect configure -state normal
		$fPortMonitor.btnStartStopMonitoring configure -state normal
		$fTabs state !disabled
		portMonitorRefreshStatus
		allPacketCaptureRefreshStatus
	}
	update
}

proc sessionsDownload {listboxSessions measurementStatusLabel btnConnect btnDownload} {
	variable paragon
	variable downloadedDataFolderPrefix
	variable cxDiagMode
	
	cxLogDebug "[info level 0]"
	set index [$listboxSessions curselection]
	set folder [$listboxSessions get $index]
	#puts "download $index $folder"
	
	$measurementStatusLabel configure -text "Download in progress..." -fg "green"
	statusUpdate "Download in progress..." "green"
	$btnConnect configure -state disabled
	$btnDownload configure -state disabled
	CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode
	dd writeDebugInfoFile
	dd downloadData $folder
	dd destroy
	$measurementStatusLabel configure -text "Download complete" -fg "black"
	statusUpdate "Download in progress... done" "black"
	$btnConnect configure -state normal
	$btnDownload configure -state normal

}

proc sessionsClear {} {
	variable fSessions
	cxLogDebug "[info level 0]"
	
	$fSessions.lbSessionList delete 0 99
	$fSessions.lblStatus configure -text ""
}

proc sessionsList {listboxSessions lblStat btnDownload} {
	variable paragon
	variable lblStatus
	variable fSessions
	variable fIp
	variable fTabs
	variable btnGetSessionList

	cxLogDebug "[info level 0]"

	sessionsClear
	$btnDownload configure -state disabled
	$btnGetSessionList configure -state disabled
	$fIp.btnConnect configure -state disabled
	$fTabs state disabled

	set okToDownload TRUE

	statusUpdate "Getting list of sessions..." "green"
	# It looks as though we can't tell whether any given session is associated
	# with a running measurement. So, we can't allow any download while
	# measurements are running.
	if {[paragon anyMeasurementRunning]} {
		statusUpdate "All measurement must be stopped before downloading" "red"
		set okToDownload FALSE
		$btnDownload configure -state disabled
	} else {
		statusUpdate "Getting list of sessions... No measurements running" "green"
		#$btnDownload configure -state normal
	}
	#$lblStat configure -text $status -fg $statusFg
	set folders [paragon getAllSessionFolders]
	set sessions [getLastSessions $folders 20]

	$listboxSessions delete 0 99
	if {[llength $sessions] == 0} {
		$listboxSessions insert 0 "No sessions on instrument"
		set okToDownload FALSE
	} else {
		foreach session $sessions {
			$listboxSessions insert end $session		
		}
		$listboxSessions selection set 0
	}
	$btnGetSessionList configure -state normal
	$fIp.btnConnect configure -state normal
	if {$okToDownload} {
		$btnDownload configure -state normal
	} else {
		$btnDownload configure -state disabled
	}
	if {$okToDownload} {
#		statusUpdate "Getting list of sessions... Done" "black"
	}
	$fTabs state !disabled

}

proc cbIpAddress {content connectButton labelSn labelStatus} {
	#puts "cbIpAddress: $content"
	cxLogDebug "[info level 0]"
	if {$content == ""} {
		$connectButton configure -state disabled
	} else {
		$connectButton configure -state normal
		#$labelSn configure -text ""
		#$labelStatus configure -text ""
	}
	return True
}

proc allDone {} {
	exit
}

proc comboPortSelectionChanged {} {
	variable fOpticalModuleInfo
	variable paragon
	set f $fOpticalModuleInfo
	cxLogDebug "[info level 0]"
	
	opticalModuleSelectedEepromContentsClear
	$f.btnDownloadOpticalModuleInfo configure -state disabled
	$fOpticalModuleInfo.lfEepromDetails.btnGetSelectedEepromContents configure -state disabled
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state disabled
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state disabled
	$f.lblGpioStatus configure -text ""
	$f.comboIfSelect configure -state disable
	update	
}

proc comboIfSelectionChanged {} {
	variable fOpticalModuleInfo
	cxLogDebug "[info level 0]"

	set f $fOpticalModuleInfo
	#puts "comboIfSelectChanged"
	
	opticalModuleSelectedEepromContentsClear
	$f.btnDownloadOpticalModuleInfo configure -state disabled
	$fOpticalModuleInfo.lfEepromDetails.btnGetSelectedEepromContents configure -state disabled
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state disabled
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state disabled
	$f.lfEepromDetails.btnStartStopPowerMeasurement configure -state disabled
	$f.lblGpioStatus configure -text ""
	update

	set selectedIf [$f.comboIfSelect get]
	set selectedPort [$f.comboPortSelect get]
	
	if {[info commands $selectedIf] != ""} {
		#puts [info commands p]
		$selectedIf destroy
	}

	CxInstrumentOpticalModule create $selectedIf $selectedPort $selectedIf
	set gpio [$selectedIf readGpio]
	if {$gpio == 0} {
		$f.lblGpioStatus configure -text "Module is inserted"
	} else {
		$f.lblGpioStatus configure -text "Module does not appear to be inserted"		
	}

	set eepromPages [$selectedIf getEepromPages]
	cxLogDebug "comboIfSelectionChanged: Pages - $eepromPages"
	$f.lfEepromWrite.comboPage configure -values $eepromPages
	if {$eepromPages != ""} {
		$f.lfEepromWrite.comboPage current 0
	}
	
	
	# Note Sfp1G is enabled here but the factory commands only support =this from neo drop#1 onwards
	# TODO: Make Sfp1G functionality dependent on the version - better would be to autodetect.
#	if {[$f.comboIfSelect get] == "SfpPlus" || [$f.comboIfSelect get] == "Sfp28" \
#			|| [$f.comboIfSelect get] == "QsfpPlus" || [$f.comboIfSelect get] == "Qsfp28" \
#			|| [$f.comboIfSelect get] == "Sfp1G" } {}
	if {[$selectedIf isRegisterAccessSupported]} {
		$f.btnDownloadOpticalModuleInfo configure -state normal
		$fOpticalModuleInfo.lfEepromDetails.btnGetSelectedEepromContents configure -state normal
		$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
		$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal
		$f.lfEepromDetails.btnStartStopPowerMeasurement configure -state normal
	}

	$selectedIf destroy
	update
}

proc nbTabsTabChanged {} {
	variable fTabs
	variable fAllPacketCapture
	variable fPortMonitor
	variable lblStatus

	cxLogDebug "[info level 0]"
	
	$lblStatus configure -text ""
	update

	#puts "Tab changed..."
	set selectedTab [$fTabs select]
	#puts $selectedTab
	#set buttonState "normal"
	if {$selectedTab == $fAllPacketCapture} {
		allPacketCaptureRefreshStatus
		#if {[info commands paragon] != ""} {
		#	set linkState [paragon port2 getLinkStatus]
		#	if {$linkState == "Link"} {
		#		set linkState "Link is up"
		#		set fg "green"
		#	} else {
		#		set linkState "Link is down"
		#		set fg "red"
		#		set buttonState "disabled"
		#	}
		#	$fAllPacketCapture.lblPort2Link configure -text $linkState -fg $fg
		#	
		#	set measurementsRunning [paragon anyMeasurementRunning]
		#	if {!$measurementsRunning} {
		#		set measState "No measurements running"
		#		set fg "green"
		#	} else {
		#		set measState "Measurements are running"
		#		set fg "red"
		#		set buttonState "disabled"
		#	}
		#	$fAllPacketCapture.btnStartCapture configure -state $buttonState
		#	$fAllPacketCapture.lblStatus configure -text $measState -fg $fg
		#	update
		#}
	} elseif {$selectedTab == $fPortMonitor} {
		if {[info commands paragon] != ""} {
			set p1If [string toupper [paragon port1 getSelectedInterface]]
			set p2If [string toupper [paragon port2 getSelectedInterface]]
			$fPortMonitor.lblPort1SelectedInterface configure -text $p1If
			$fPortMonitor.lblPort2SelectedInterface configure -text $p2If
			$fPortMonitor.lfItemSelection.cbRxPower select
			if {$p1If == "QSFP28" || $p1If == "QSFPPLUS" || $p2If == "QSFP28" || $p2If == "QSFPPLUS"} {
				$fPortMonitor.lfItemSelection.cbBips select				
			}
			
			# Check what is available. If neither interface supports monitoring
			# then disable the start button
			set availableMonitors [getAvailableMonitorsForSelectedInterface]
			set p1 [dict get $availableMonitors "Port 1"]
			set p2 [dict get $availableMonitors "Port 2"]
			if {[lindex $p1 0] == "NONE" && [lindex $p2 0] == "NONE"} {
				$fPortMonitor.btnStartStopMonitoring configure -state disabled
			} else {
				$fPortMonitor.btnStartStopMonitoring configure -state normal				
			}
			portMonitorConfigureSelection $p1 $p2
			
			update
		}
	} else {
		$fAllPacketCapture.lblPort2Link configure -text "" -fg black
		$fAllPacketCapture.lblStatus configure -text "" -fg black
		$fAllPacketCapture.btnStartCapture configure -state disabled		
	}
}

proc debugInfoDownload {} {
	variable paragon
	variable fIp
	variable fTabs
	variable fDebugInfo
	variable downloadedDataFolderPrefix
	variable cxDiagMode

	cxLogDebug "[info level 0]"
	
#	$measurementStatusLabel configure -text "Download in progress..." -fg "green"
	$fIp.btnConnect configure -state disabled
	$fTabs state disabled
	$fDebugInfo.btnGetDebugInfo configure -state disabled
	$fDebugInfo.btnDownloadDebugInfo configure -state disabled
	statusUpdate "Downloading data..." "green"
	
	CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode
	dd writeDebugInfoFile
	dd downloadLogs
	dd destroy
	#$measurementStatusLabel configure -text "Download complete" -fg "black"
	$fIp.btnConnect configure -state normal
	$fDebugInfo.btnGetDebugInfo configure -state normal
	$fDebugInfo.btnDownloadDebugInfo configure -state normal
	$fTabs state !disabled
	statusUpdate "Downloading data... done" "black"
}

proc allPacketCaptureStart {f} {
	variable paragon
	variable fIp
	variable fTabs
	cxLogDebug "[info level 0]"

	$fIp.btnConnect configure -state disabled
	$f.btnStartCapture configure -state disabled
	$f.btnPfv configure -state disabled
	$f.btnDownload configure -state disabled
	$f.lblCaptureSessionName configure -text ""
	$fTabs state disabled

	statusUpdate "Running all packet capture..." "green"
	set durationS [$f.entryCaptureDuration get]
	set currentPreset [paragon getCurrentPreset]
	if {$currentPreset != "PTP 1588"} {
		$f.lblStatus configure -text " "
		$f.lblStatus configure -text "Changing preset to PTP..."
		paragon selectPreset {"PTP 1588"}
	}
	# Start the capture
	paragon startMseCapture
	$f.lblStatus configure -text " "
	# Clear the filters
	paragon clearCaptureFilters
	set t [expr $durationS*1000]
	$f.lblStatus configure -text "Capturing... [expr $t / 1000] seconds remaining" -fg "green"
	update
	while {$t > 0} {
		after 1000
		set t [expr $t - 1000]
		$f.lblStatus configure -text "Capturing...[expr $t / 1000] seconds remaining" -fg "green"
		update
	}
	paragon stopMseCapture
	$f.lblStatus configure -text " "
	$f.lblStatus configure -text "No measurements running"
	$fIp.btnConnect configure -state normal
	$f.btnStartCapture configure -state normal
	
	set sessionName [paragon getLastSessionsFolder]
	$f.lblCaptureSessionName configure -text $sessionName
	$f.btnPfv configure -state normal
	$f.btnDownload configure -state normal
	$fTabs state !disabled
	statusUpdate "Running all packet capture... done" "black"

}

proc launchPfv {f} {
	variable paragon
	cxLogDebug "[info level 0]"
	paragon pfvShow
	return
}
proc allPacketCaptureDownloadSession {f} {
	variable paragon
	variable downloadedDataFolderPrefix
	variable lblStatus
	variable cxDiagMode
	cxLogDebug "[info level 0]"
	
	CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode
	statusUpdate "Downloading logs... this may take some time" "green"
	dd writeDebugInfoFile
	statusUpdate "Downloading session files... this may take some time" "green"
	dd downloadData [$f.lblCaptureSessionName cget -text]
	dd destroy
	statusUpdate "Downloading complete." "green"
	return
}

proc allPacketCaptureRefreshStatus {} {
	variable fAllPacketCapture
    variable fTabs
	cxLogDebug "[info level 0]"

	set buttonState "normal"

	if {[info commands paragon] != ""} {
		$fAllPacketCapture.btnStartCapture configure -state disabled
		$fAllPacketCapture.lblPort2Link configure -text "" -fg black
		$fAllPacketCapture.lblStatus configure -text "" -fg black
		update
		
		set linkState [paragon port2 getLinkStatus]
		if {$linkState == "Link"} {
			set linkState "Link is up"
			set fg "green"
		} else {
			set linkState "Link is down"
			set fg "red"
			set buttonState "disabled"
		}
		$fAllPacketCapture.lblPort2Link configure -text $linkState -fg $fg
			
		set measurementsRunning [paragon anyMeasurementRunning]
		if {!$measurementsRunning} {
			set measState "No measurements running"
			set fg "green"
		} else {
			set measState "Measurements are running"
			set fg "red"
			set buttonState "disabled"
		}
		$fAllPacketCapture.btnStartCapture configure -state $buttonState
		$fAllPacketCapture.lblStatus configure -text $measState -fg $fg

        set ptpFitted [paragon isPtpOptionFitted]
        if {!$ptpFitted} {
            #$fAllPacketCapture configure -state disabled
            $fTabs tab $fAllPacketCapture -state disabled
        }
		update
	}
	
}

proc allPacketCaptureClear {} {
	variable fAllPacketCapture
	cxLogDebug "[info level 0]"
	
	$fAllPacketCapture.lblPort2Link configure -text "" -fg black
	$fAllPacketCapture.lblStatus configure -text "" -fg black
	$fAllPacketCapture.lblCaptureSessionName configure -text "" -fg black
	$fAllPacketCapture.btnStartCapture configure -state disabled
	$fAllPacketCapture.btnPfv configure -state disabled
	$fAllPacketCapture.btnDownload configure -state disabled
}

proc debugInfoClear {} {
	variable fDebugInfo
	cxLogDebug "[info level 0]"

	$fDebugInfo.lfInstrDetails.lblSwVersion configure -text ""
	$fDebugInfo.lfInstrDetails.lblOsVersion configure -text ""
	$fDebugInfo.lfInstrDetails.lblHwCapabilities configure -text ""

	set keys [$fDebugInfo.tvPortDetails children {}]
	$fDebugInfo.tvPortDetails delete $keys
	
#	set keys [$fDebugInfo.tvPortDetails children {}]
#	foreach k $keys {
#		set val ""
#		$fDebugInfo.tvPortDetails item $k -values [list "" ""]
#	}
	$fDebugInfo.btnDownloadDebugInfo configure -state disabled
	
	
}

proc debugInfoGet {lblSwVersion lblOsVersion lblHwCapabilities tvPorts btnDownloadDebugInfo} {
	variable fDebugInfo
	variable lblStatus
	cxLogDebug "[info level 0]"

	set f $fDebugInfo
	debugInfoClear

	puts "getDebugInfo..."
	statusUpdate "Reading debug data..." green
	$lblSwVersion configure -text "[paragon getSoftwareVersion]"
	$lblOsVersion configure -text "[paragon getOsVersion]"
	$lblHwCapabilities configure -text "[paragon getCapabilities]"
	
	set port1Settings [paragon port1 getEthernetSettings]
	set port2Settings [paragon port2 getEthernetSettings]
	set port1Interfaces [dict get $port1Settings "Interfaces"]
	set port2Interfaces [dict get $port2Settings "Interfaces"]

	#puts $port1Interfaces
	set keysIfs [dict keys $port1Interfaces]

	foreach key $keysIfs {
		$f.tvPortDetails insert {} end -id $key -text $key
		set ifDetails [dict get $port1Interfaces $key]
		#puts $ifDetails
		#set keysIfDetails [dict keys $ifDetails]
		set keysIfDetails {"InterfaceType" "Fitted" "LineRate" "Reach" "Manufacturer" "PartNumber" "SerialNumber"}
		set detailsKeys {"Reach" "Manufacturer" "PartNumber" "SerialNumber"}
		#puts $keysIfDetails
		foreach subkey $keysIfDetails {
			$f.tvPortDetails insert $key end -id $key$subkey -text $subkey
			set p1Info [dict get $port1Interfaces $key]
			if {$subkey in $detailsKeys} {
				if {[dict exists $port1Interfaces $key "Details"]} {
					set p1Info [dict get $port1Interfaces $key "Details"]
				} else {
					set p1Info "NA NA"
				}
			}
			if {[dict exists $p1Info $subkey]} {
				set p1Val [dict get $p1Info $subkey]
			} else {
				set p1Val "NA"
			}
			set p2Info [dict get $port2Interfaces $key]
			if {$subkey in $detailsKeys} {
				if {[dict exists $port2Interfaces $key "Details"]} {
					set p2Info [dict get $port2Interfaces $key "Details"]
				} else {
					set p2Info "NA NA"
				}


				#set p2Info [dict get $port2Interfaces $key "Details"]				
			}
			if {[dict exists $p2Info $subkey]} {
				set p2Val [dict get $p2Info $subkey]
			} else {
				set p2Val "NA"
			}
			$f.tvPortDetails item $key$subkey -values [list $p1Val $p2Val]
		}
		$btnDownloadDebugInfo configure -state normal
		statusUpdate "Reading debug data... done" black
	}
	

	#set selectedIfP1 [paragon port1 getSelectedInterface]
	#set selectedIfP2 [paragon port2 getSelectedInterface]
	#set ifInfoP1 [paragon port1 getSelectedInterfaceDetails]
	#set ifInfoP2 [paragon port2 getSelectedInterfaceDetails]
	#puts $ifInfoP1
	#puts $ifInfoP1
	#if {[dict exists $ifInfoP1 "Details"]} {
	#	set ifDetailsP1 [dict get $ifInfoP1 "Details"]
	#	puts $ifDetailsP1
	#}
	#if {[dict exists $ifInfoP2 "Details"]} {
	#	set ifDetailsP2 [dict get $ifInfoP2 "Details"]
	#	puts $ifDetailsP2
	#}
	
	#set keys [$tvPorts children {}]
	#foreach k $keys {
	#	set p1Val ""
	#	set p2Val ""
	#	if {[dict exists $ifInfoP1 $k]} {
	#		set p1Val [dict get $ifInfoP1 $k]
	#	} else {
	#		if {[dict exists $ifInfoP1 "Details"] && [dict exists $ifDetailsP1 $k]} {
	#			set p1Val [dict get $ifDetailsP1 $k]
	#		}
	#	}
	#	if {[dict exists $ifInfoP2 $k]} {
	#		set p2Val [dict get $ifInfoP2 $k]
	#	} else {
	#		if {[dict exists $ifInfoP2 "Details"] && [dict exists $ifDetailsP2 $k]} {
	#			set p2Val [dict get $ifDetailsP2 $k]
	#		}
	#	}
	#	$tvPorts item $k -values [list $p1Val $p2Val]
	#}
}

proc opticalModuleSelectedEepromContentsClear {} {
	variable fOpticalModuleInfo
	cxLogDebug "[info level 0]"

	set f $fOpticalModuleInfo.lfEepromDetails
	
	#puts [winfo children $f]
	$f.lblType configure -text ""
	$f.lblEncoding configure -text ""
	$f.lblBrNominal configure -text ""
	$f.lblExtCompliance configure -text ""

}

proc opticalModuleInfoClear {} {
	variable fOpticalModuleInfo
	cxLogDebug "[info level 0]"
	
	set f $fOpticalModuleInfo
	
#	set keys [$fOpticalModuleInfo.tvPortDetails children {}]
#	$fOpticalModuleInfo.tvPortDetails delete $keys
	#foreach k $keys {
	#	set val ""
	#	$fOpticalModuleInfo.tvPortDetails item $k -values [list "" ""]
	#}
	$f.btnDownloadOpticalModuleInfo configure -state disabled	
	$f.btnGetOpticalModuleInterfaces configure -state disabled	
	$f.lfEepromDetails.btnGetSelectedEepromContents configure -state disabled
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state disabled
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state disabled
	$f.lblStatus configure -text ""
	#$f.lblGpioStatus configure -text ""
	
	set keys [$f.lfEepromDetails.tvPowerDetails children {}]
	$f.lfEepromDetails.tvPowerDetails delete $keys
	
	#$f.lfEepromDetails.btnStartStopPowerMeasurement configure -state disabled

}




proc opticalModuleInfoGet {} {
	variable fOpticalModuleInfo
	variable fOpticalModuleEepromLabelFrame
	variable fTabs
	variable fIp
	cxLogDebug "[info level 0]"
	#puts "opticalModuleInfoGet..."

	set f $fOpticalModuleInfo

	opticalModuleInfoClear
	$fTabs state disabled
	$fOpticalModuleEepromLabelFrame.btnStartStopPowerMeasurement configure -state disabled
	$fIp.btnConnect configure -state disabled
	
	set port1Settings [paragon port1 getEthernetSettings]
	set port2Settings [paragon port2 getEthernetSettings]
	set port1Interfaces [dict get $port1Settings "Interfaces"]
	set port2Interfaces [dict get $port2Settings "Interfaces"]

	#puts $port1Interfaces
	set keysIfs [dict keys $port1Interfaces]
	
	#puts $fIfs
	set nonOpticalIfs {"Rj45_LowRate" "Rj45_1G"}
	
	foreach nonOpticalIf $nonOpticalIfs {
		set idx [lsearch $keysIfs $nonOpticalIf]
		set keysIfs [lreplace $keysIfs $idx $idx]
	}
	
	#if {[dict exists $keysIfs "Rj45_1G"]} {
	#	set keysIfs [dict remove $keysIfs "Rj45_1G"]	
	#}
	
	#puts $keysIfs
	#set keysIfDetails [dict keys $details]
	set selectedPort [$f.comboPortSelect get]
	if {$selectedPort == "Port 1"} {
		$f.lblSelectedInterface configure -text [paragon port1 getSelectedInterface]		
	} else {
		$f.lblSelectedInterface configure -text [paragon port2 getSelectedInterface]				
	}

	##### HACK HACK HACK
	# EPPROM reads aren't support for all interfaces and so we need to filter
	$f.comboIfSelect configure -values ""
	if {$selectedPort == "Port 1"} {
#		$f.comboIfSelect configure -values [paragon port1 getFittedInterfaces]
		$f.comboIfSelect configure -values $keysIfs
	} elseif {$selectedPort == "Port 2"} {
#		$f.comboIfSelect configure -values [paragon port2 getFittedInterfaces]
		$f.comboIfSelect configure -values $keysIfs
	}
	$f.comboIfSelect current 0

	##### HACK HACK HACK
	# EPPROM reads aren't support for all interfaces
	# Enable download for only those that are supported
	if {[$f.comboIfSelect get] == "sfpplus" || [$f.comboIfSelect get] == "sfp28"} {
		$f.btnDownloadOpticalModuleInfo configure -state normal
		$fOpticalModuleInfo.lfEepromDetails.btnGetSelectedEepromContents configure -state normal
		$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
		$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal
	}
	$f.comboIfSelect configure -state readonly
	$f.btnGetOpticalModuleInterfaces configure -state normal
	#$fOpticalModulepromLabelFrame.btnStartStopPowerMeasurement configure -state normal
	$fTabs state !disabled
	$fIp.btnConnect configure -state normal

}

proc opticalModuleInfoDownload {} {
	variable paragon
	variable fIp
	variable fOpticalModuleInfo
	variable fTabs
	variable downloadedDataFolderPrefix
	variable cxDiagMode
	
	cxLogDebug "[info level 0]"

	set f $fOpticalModuleInfo
	
	statusUpdate "Reading optical module registers..." green
#	$measurementStatusLabel configure -text "Download in progress..." -fg "green"
	$fTabs state disabled
	$fIp.btnConnect configure -state disabled
	$f.btnDownloadOpticalModuleInfo configure -state disabled
	$f.btnGetOpticalModuleInterfaces configure -state disabled
	$f.lfEepromDetails.btnGetSelectedEepromContents configure -state disabled
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state disabled
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state disabled
	$f.lfEepromDetails.btnStartStopPowerMeasurement configure -state disabled
	
	set portName [$fOpticalModuleInfo.comboPortSelect get]
	set interface [$fOpticalModuleInfo.comboIfSelect get]
	
	CxInstrumentOpticalModule create $interface $portName $interface
	#puts [$interface readRegister 0]

	
	CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode

	$fOpticalModuleInfo.lblStatus configure -text "Writing debug info file..."
	update
	dd writeDebugInfoFile
	$fOpticalModuleInfo.lblStatus configure -text "Reading $interface registers..."
	update
	dd writeOpticalModuleInfo $interface $fOpticalModuleInfo.lblStatus
	$fOpticalModuleInfo.lblStatus configure -text "Downloading logs... this may take some time"
	update
	dd downloadLogs
	$fOpticalModuleInfo.lblStatus configure -text "Done"
	update
	dd destroy
	#$measurementStatusLabel configure -text "Download complete" -fg "black"

	$interface destroy

	$fTabs state !disabled
	$fIp.btnConnect configure -state normal
	$f.btnDownloadOpticalModuleInfo configure -state normal
	$f.btnGetOpticalModuleInterfaces configure -state normal
	$f.lfEepromDetails.btnGetSelectedEepromContents configure -state normal
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal

	statusUpdate "Reading optical module registers... done" black
}

proc opticalModuleSelectedEepromContentsRead {} {
	variable fOpticalModuleInfo
	variable fOpticalModuleEepromLabelFrame
	cxLogDebug "[info level 0]"
	
	set f $fOpticalModuleEepromLabelFrame
	
	opticalModuleInfoClear
	$f.btnStartStopPowerMeasurement configure -state disabled
	
	set portName [$fOpticalModuleInfo.comboPortSelect get]
	set interface [$fOpticalModuleInfo.comboIfSelect get]
	
	CxInstrumentOpticalModule create $interface $portName $interface
	$f.lblType configure -text [$interface getEepromField Type]
	update
	$f.lblEncoding configure -text [$interface getEepromField Encoding]
	update
	#$f.lblBrNominal configure -text [$interface getEepromField BrNominal]
	$f.lblBrNominal configure -text [$interface getNominalRate]
	update
	$f.lblExtCompliance configure -text [$interface getEepromField ExtCompliance]
	update
	
	# This is old code. Comment out for the moment
	set keys [$f.tvPowerDetails children {}]
	$f.tvPowerDetails delete $keys

#	set powerList [$interface getPowers]
#	set numPowers [llength $powerList]
#	set powers ""
#	for {set i 0} {$i < $numPowers} {incr i} {
#		set powers [lindex $powerList $i]			
#		set rx [format %0.4f [lindex $powers 0]]
#		set tx [format %0.4f [lindex $powers 1]]
#		$f.tvPowerDetails insert {} end -id "Power$i" -text "Power [expr $i+1] (mW)"
#		$f.tvPowerDetails item "Power$i" -values [list $rx $tx]
#		update
#	}

	$fOpticalModuleInfo.btnDownloadOpticalModuleInfo configure -state normal
	$fOpticalModuleInfo.btnGetOpticalModuleInterfaces configure -state normal	
	$f.btnGetSelectedEepromContents configure -state normal
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal
	$f.btnStartStopPowerMeasurement configure -state normal

	$interface destroy
}



proc opticalModuleSelectedEepromAddressRead {} {
	variable fOpticalModuleInfo
	variable fOpticalModuleEepromLabelFrame
	variable fOpticalModuleEepromWriteLabelFrame
	cxLogDebug "[info level 0]"
	statusUpdate "Reading module register..." green
	
	set f $fOpticalModuleInfo
	
	opticalModuleInfoClear
	
	set portName [$f.comboPortSelect get]
	set interface [$f.comboIfSelect get]
	set page [$f.lfEepromWrite.comboPage get]
	set address [$f.lfEepromWrite.entryAddr get]

	if {$address == "" || $page == ""} {
		statusUpdate "Address not complete - need page and register address" red
	} else {
		CxInstrumentOpticalModule create $interface $portName $interface
		set name ""
		set regRange [$interface getRegisterRange $page]
		set low [dict get $regRange "first"]
		set upp [dict get $regRange "last"]
		if {$address < $low || $address > $upp} {
			statusUpdate "Address must be between $low and $upp" red			
			$interface destroy
		} else {
			set reg [$interface readRegister $page $address]
			set data [lindex $reg 0]
			set name [lindex $reg 1]
			$f.lfEepromWrite.entryData delete 0 end
			set d [format "%X" $data]
			set d [string toupper $d]
			$f.lfEepromWrite.entryData insert 0 $d			
			$interface destroy
			
			if {$data == -1} {
				statusUpdate "Reading module register...ERROR: Page $page not supported" red							
			} else {
				if {$name != ""} {
					statusUpdate "Reading module register ($name)...done" green
				} else {
					statusUpdate "Reading module register...done" green			
				}
			}
		}

	}
	$f.btnDownloadOpticalModuleInfo configure -state normal
	$f.btnGetOpticalModuleInterfaces configure -state normal	
	$f.lfEepromDetails.btnGetSelectedEepromContents configure -state normal
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal
}

proc opticalModuleSelectedEepromAddressWrite {} {
	variable fOpticalModuleInfo
	variable fOpticalModuleEepromLabelFrame
	variable fOpticalModuleEepromWriteLabelFrame
	cxLogDebug "[info level 0]"
	statusUpdate "Writing module register..." green
	
	set f $fOpticalModuleInfo
	
	opticalModuleInfoClear
	
	set portName [$f.comboPortSelect get]
	set interface [$f.comboIfSelect get]
	set page [$f.lfEepromWrite.comboPage get]
	set address [$f.lfEepromWrite.entryAddr get]
	set dataHex [$f.lfEepromWrite.entryData get]

	cxLogDebug "portName:$portName; interface:$interface; page:$page; address:$address; dataHex:$dataHex"
		
	if {$address == "" || $page == "" || $dataHex == ""} {
		statusUpdate "Invalid address or data" red
	} elseif {![string is xdigit $dataHex]} {
		statusUpdate "Invalid data (must be hex)" red				
	} elseif {[expr 0x$dataHex] > 256} {
		statusUpdate "Invalid data (must be less than 0xFF)" red		
	} else {
		CxInstrumentOpticalModule create $interface $portName $interface
		set name ""
		set regRange [$interface getRegisterRange $page]
		set low [dict get $regRange "first"]
		set upp [dict get $regRange "last"]
		if {$address < $low || $address > $upp} {
			statusUpdate "Address must be between $low and $upp" red			
			$interface destroy
		} else {
			cxLogDebug "dataHex:$dataHex"
			set dataDec [expr 0x$dataHex]
			set regWritten [$interface writeRegister $page $address $dataDec]
			set regWrittenData [lindex $regWritten 0]
			
			if {$regWrittenData == -1} {
				statusUpdate "Writing module register...ERROR: Page $page not supported" red							
			} else {
				set reg [$interface readRegister $page $address]
				set data [lindex $reg 0]
				set name [lindex $reg 1]
		
				#set dataHex [string tolower $dataHex]
				#set data [format %x $data]
				if {$name == ""} {
					set statusStr "Writing module register ($page:$address)..."
				} else {
					set statusStr "Writing module register ($name - $page:$address)..."			
				}

				if {$data != $dataDec} {
					statusUpdate "$statusStr FAILED" red						
				} else {
					statusUpdate "$statusStr done" green
				}
			}
		}
	}
	$interface destroy
	$f.btnDownloadOpticalModuleInfo configure -state normal
	$f.btnGetOpticalModuleInterfaces configure -state normal	
	$f.lfEepromDetails.btnGetSelectedEepromContents configure -state normal
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
	$f.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal
}




proc opticalModuleStopPowerMeasurement {} {
	variable fOpticalModuleEepromLabelFrame
	variable powerMeasurementStop
	cxLogDebug "[info level 0]"

	set f $fOpticalModuleEepromLabelFrame
	set powerMeasurementStop TRUE

}

# This is incomplete - it probably needs some asynchronous processing that I don't know how to do yet.
proc opticalModuleStartPowerMeasurement {} {
	variable fOpticalModuleInfo
	variable fOpticalModuleEepromLabelFrame
	variable fTabs
	variable fIp
	variable powerMeasurementRunning
	variable powerMeasurementStop
	variable paragon
	variable downloadedDataFolderPrefix
	cxLogDebug "[info level 0]"


	set f $fOpticalModuleEepromLabelFrame
	$fTabs state disabled
	$fIp.btnConnect configure -state disabled
	
	$f.btnStartStopPowerMeasurement configure -text "Stop Reading Power Levels" -state normal -command opticalModuleStopPowerMeasurement -fg "red"
	set powerMeasurementStop FALSE
	update
	
	if {!$powerMeasurementRunning} {
		variable cxDiagMode

		# Start the power measurement
		opticalModuleInfoClear
	
		set portName [$fOpticalModuleInfo.comboPortSelect get]
		set interface [$fOpticalModuleInfo.comboIfSelect get]
	
		CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode

		set fid [dd openPowerMeasurementFile $portName $interface]
		#set filename "[dd getRoot]/opticalModulePower_${portName}_${interface}_$date.csv"
		#set fid [open $filename "w"]

		puts $fid "Serial Number, Instrument Model, SW Version, Transceiver Type, Port"
		puts $fid "[paragon getSerialNumber], [paragon getModel], [paragon getSoftwareVersion], $interface, $portName"
		puts $fid ""

		CxInstrumentOpticalModule create $interface $portName $interface
		set keys [$f.tvPowerDetails children {}]
		$f.tvPowerDetails delete $keys

		# Do one read to initialise the treeview, then update
		set powerList [$interface getPowers]
		set numPowers [llength $powerList]

		if {$numPowers == 1} {
			puts $fid "Rx, Tx"			
		} elseif {$numPowers == 4} {
			puts $fid "Rx1, Tx1, Rx2, Tx2, Rx3, Tx3, Rx4, Tx4"						
		} else {
			error "Unexpected number of power readings ($numPowers)"
		}
		
		set powers ""
		for {set i 0} {$i < $numPowers} {incr i} {
				set powers [lindex $powerList $i]			
				set rx [format %0.4f [lindex $powers 0]]
				set tx [format %0.4f [lindex $powers 1]]
				$f.tvPowerDetails insert {} end -id "Power$i" -text "Power [expr $i+1] (mW)"
				$f.tvPowerDetails item "Power$i" -values [list $rx $tx]
				puts -nonewline $fid "$rx, $tx,"
				update
		}
		
		puts $fid ""
		while {!$powerMeasurementStop} {
			set powerList [$interface getPowers]
			set numPowers [llength $powerList]
			set powers ""
			for {set i 0} {$i < $numPowers} {incr i} {
				set powers [lindex $powerList $i]			
				set rx [format %0.4f [lindex $powers 0]]
				set tx [format %0.4f [lindex $powers 1]]
			
				#$f.tvPowerDetails insert {} end -id "Power$i" -text "Power [expr $i+1] (mW)"
				$f.tvPowerDetails item "Power$i" -values [list $rx $tx]
				puts -nonewline $fid "$rx, $tx,"
				update
			}
			puts $fid ""
		}

		$f.btnStartStopPowerMeasurement configure -text "Start Reading Power Levels" -state normal -command opticalModuleStartPowerMeasurement -fg "black"
		set powerMeasurementStop TRUE
		set powerMeasurementRunning FALSE
		
		close $fid
		dd destroy
	}

	$fOpticalModuleInfo.btnDownloadOpticalModuleInfo configure -state normal
	$fOpticalModuleInfo.btnGetOpticalModuleInterfaces configure -state normal	
	$f.btnGetSelectedEepromContents configure -state normal
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressRead configure -state normal
	$fOpticalModuleInfo.lfEepromWrite.btnOpticalModuleSelectedEepromAddressWrite configure -state normal
	$fTabs state !disabled
	$fIp.btnConnect configure -state normal

	$interface destroy

}

proc portMonitorConfigureSelection {p1AvailableMonitors p2AvailableMonitors} {
	variable fPortMonitorItemSelectionLabelFrame
	variable fPortMonitor
	variable portMonitorRxPowerSelected
	variable portMonitorRxPowerAlarmsSelected
	variable portMonitorTxPowerSelected
	variable portMonitorLosSelected
	variable portMonitorBipsSelected
	cxLogDebug "[info level 0]"

	set f $fPortMonitorItemSelectionLabelFrame

	if {"RxPower" in $p1AvailableMonitors || "RxPower" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbRxPower configure -state normal				
	} else {
		$fPortMonitor.lfItemSelection.cbRxPower deselect
		$fPortMonitor.lfItemSelection.cbRxPower configure -state disabled
	}
	if {"BIPS" in $p1AvailableMonitors || "BIPS" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbBips configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbBips deselect
		$fPortMonitor.lfItemSelection.cbBips configure -state disabled
	}
	if {"LOS" in $p1AvailableMonitors || "LOS" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbLos configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbLos deselect
		$fPortMonitor.lfItemSelection.cbLos configure -state disabled
	}
	if {"TxPower" in $p1AvailableMonitors || "TxPower" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbTxPower configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbTxPower deselect
		$fPortMonitor.lfItemSelection.cbTxPower configure -state disabled
	}
	if {"RxPowerAlarms" in $p1AvailableMonitors || "RxPowerAlarms" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbRxPowerAlarms configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbRxPowerAlarms deselect
		$fPortMonitor.lfItemSelection.cbRxPowerAlarms configure -state disabled
	}
	if {"FECcorr" in $p1AvailableMonitors || "FECcorr" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbFecCorr configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbFecCorr deselect
		$fPortMonitor.lfItemSelection.cbFecCorr configure -state disabled
	}
	if {"FECuncorr" in $p1AvailableMonitors || "FECuncorr" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbFecUncorr configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbFecUncorr deselect
		$fPortMonitor.lfItemSelection.cbFecUncorr configure -state disabled
	}
	if {"FEClock" in $p1AvailableMonitors || "FEClock" in $p2AvailableMonitors} {
		$fPortMonitor.lfItemSelection.cbFecLock configure -state normal
	} else {
		$fPortMonitor.lfItemSelection.cbFecLock deselect
		$fPortMonitor.lfItemSelection.cbFecLock configure -state disabled
	}
	
}

proc portMonitorItemsClear {} {
	variable fPortMonitor
	cxLogDebug "[info level 0]"

	set f $fPortMonitor

	set keys [$f.tvMonitorData children {}]
	$f.tvMonitorData delete $keys
	
}

proc portMonitorRefreshStatus {} {
	variable fPortMonitor
	variable cxDiagMode
	cxLogDebug "[info level 0]"
	
	# Nothing to do if we are in basic mode
	if {$cxDiagMode == "BASIC"} {
		return
	}
	
	if {[info commands paragon] != ""} {
		set p1If [string toupper [paragon port1 getSelectedInterface]]
		set p2If [string toupper [paragon port2 getSelectedInterface]]
		$fPortMonitor.lblPort1SelectedInterface configure -text $p1If
		$fPortMonitor.lblPort2SelectedInterface configure -text $p2If
		$fPortMonitor.lfItemSelection.cbRxPower select
		if {$p1If == "QSFP28" || $p1If == "QSFPPLUS" || $p2If == "QSFP28" || $p2If == "QSFPPLUS"} {
			$fPortMonitor.lfItemSelection.cbBips select				
		}
		if {$p1If == "QSFP28" || $p1If == "SFP28" || $p1If == "CFP4"} {
			set p1Fec [paragon port1 getFecState $p1If]
			$fPortMonitor.lblPort1SelectedInterfaceFec configure -text $p1Fec
		} else {
			$fPortMonitor.lblPort1SelectedInterfaceFec configure -text NA			
		}
		if {$p2If == "QSFP28" || $p2If == "SFP28" || $p2If == "CFP4"} {
			set p2Fec [paragon port2 getFecState $p2If]
			$fPortMonitor.lblPort2SelectedInterfaceFec configure -text $p2Fec
		} else {
			$fPortMonitor.lblPort2SelectedInterfaceFec configure -text NA			
		}
			
		# Check what is available. If neither interface supports monitoring
		# then disable the start button
		set availableMonitors [getAvailableMonitorsForSelectedInterface]
		set p1 [dict get $availableMonitors "Port 1"]
		set p2 [dict get $availableMonitors "Port 2"]
		if {[lindex $p1 0] == "NONE" && [lindex $p2 0] == "NONE"} {
			$fPortMonitor.btnStartStopMonitoring configure -state disabled
		} else {
			$fPortMonitor.btnStartStopMonitoring configure -state normal				
		}
		portMonitorConfigureSelection $p1 $p2
			
		update
	}
	
}

proc portMonitorGetSelectedMonitorItems {} {
	variable fPortMonitorItemSelectionLabelFrame
	variable portMonitorRxPowerSelected
	variable portMonitorRxPowerAlarmsSelected
	variable portMonitorTxPowerSelected
	variable portMonitorLosSelected
	variable portMonitorBipsSelected
	variable portMonitorFecCorrSelected
	variable portMonitorFecUncorrSelected
	variable portMonitorFecLockSelected
	cxLogDebug "[info level 0]"

	set items [list]
	
	set f $fPortMonitorItemSelectionLabelFrame
	if {$portMonitorRxPowerSelected} {
		lappend items "RxPower"		
	}
	if {$portMonitorRxPowerAlarmsSelected} {
		lappend items "RxPowerAlarms"				
	}
	if {$portMonitorTxPowerSelected} {
		lappend items "TxPower"		
	}
	if {$portMonitorLosSelected} {
		lappend items "LOS"		
	}
	if {$portMonitorBipsSelected} {
		lappend items "BIPS"		
	}
	if {$portMonitorFecCorrSelected} {
		lappend items "FECcorr"		
	}
	if {$portMonitorFecUncorrSelected} {
		lappend items "FECuncorr"		
	}
	if {$portMonitorFecLockSelected} {
		lappend items "FEClock"		
	}
	
	return $items
}

proc portMonitorStop {} {
	variable portMonitorStop
	cxLogDebug "[info level 0]"

	set portMonitorStop TRUE
}

proc getAvailableMonitorsForSelectedInterface {} {
	set availableMonitorItems [dict create "Port 1" NONE "Port 2" NONE]
	cxLogDebug "[info level 0]"

	foreach port {"Port 1" "Port 2"} {
		if {$port == "Port 1"} {			
			set selIf [paragon port1 getSelectedInterface]
			CxInstrumentOpticalModule create "port1$selIf" $port $selIf
			dict set availableMonitors $port ["port1$selIf" getAvailableMonitorItems]
			"port1$selIf" destroy 
		} else {
			set selIf [paragon port2 getSelectedInterface]
			CxInstrumentOpticalModule create "port2$selIf" $port $selIf
			dict set availableMonitors $port ["port2$selIf" getAvailableMonitorItems]
			"port2$selIf" destroy
		}
	}	
	return $availableMonitors
}

proc portMonitorStart {} {
	variable fPortMonitorItemSelectionLabelFrame
	variable fPortMonitor
	variable fTabs
	variable fIp
	variable paragon
	variable portMonitorStop
	variable downloadedDataFolderPrefix
	variable cxDiagMode
	cxLogDebug "[info level 0]"

	portMonitorItemsClear
	set f $fPortMonitor
	$fTabs state disabled
	$fIp.btnConnect configure -state disabled
	
	$f.btnStartStopMonitoring configure -text "Stop" -state normal -command portMonitorStop -fg "red"
	set portMonitorStop FALSE
	statusUpdate "Port monitor running..." green
	update

	set selectedInterfaces [dict create "Port 1" NONE "Port 2" NONE]
	set availableMonitorItems [dict create "Port 1" NONE "Port 2" NONE]
	
	CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode

	set fid [dd openPortMonitorFile]

	puts $fid "Serial Number, Instrument Model, SW Version"
	puts $fid "[paragon getSerialNumber], [paragon getModel], [paragon getSoftwareVersion]"
	puts $fid ""

	
	foreach port {"Port 1" "Port 2"} {
		if {$port == "Port 1"} {			
			set p1SelIf [paragon port1 getSelectedInterface]
			CxInstrumentOpticalModule create "port1$p1SelIf" $port $p1SelIf
			dict set availableMonitorItems $port ["port1$p1SelIf" getAvailableMonitorItems]
			dict set selectedInterfaces $port $p1SelIf
		} else {
			set p2SelIf [paragon port2 getSelectedInterface]
			CxInstrumentOpticalModule create "port2$p2SelIf" $port $p2SelIf
			dict set availableMonitorItems $port ["port2$p2SelIf" getAvailableMonitorItems]
			dict set selectedInterfaces $port $p2SelIf			
		}
		#lappend selectedInterfaces $selectedIf
		#lappend availableMonitorItems $availableMonitors
	}

	set items [portMonitorGetSelectedMonitorItems]
	foreach item $items {
		$f.tvMonitorData insert {} end -id $item -text $item
	}

	set headerRow ""
	foreach item $items {
		foreach {portName} {"P1" "P2"} {
			set colHeader $portName.$item
			append headerRow $colHeader ","
		}
	}
	puts $fid $headerRow
	
	set loopCount 0
	while {!$portMonitorStop} {
		set loopCount [expr $loopCount + 1]
		statusUpdate "Port monitor running... ($loopCount)" green
		update
		if {$portMonitorStop} {
			break
		}
		foreach item $items {
			if {$portMonitorStop} {
				break
			}
			set monitorResult [list]
			foreach {portName portId} {"Port 1" port1 "Port 2" port2} {
				if {$portMonitorStop} {
					$f.tvMonitorData item $item -values $monitorResult
					update
					break
				}
				set columnHeader ""
				if {$item in [dict get $availableMonitorItems $portName]} {
					set interface [dict get $selectedInterfaces $portName]
					set data [$portId$interface getMonitorData $item]
					cxLogDebug $data
					puts -nonewline $fid "$data,"
					lappend monitorResult $data
				} else {
					puts -nonewline $fid ","					
				}
				
			}
			#$f.tvMonitorData item $item -values [list $data ""]
			$f.tvMonitorData item $item -values $monitorResult
		}
		puts $fid ""
		update
	}

	$f.btnStartStopMonitoring configure -text "Start" -state normal -command portMonitorStart -fg "black"
	set portMonitorStop TRUE
	$fTabs state !disabled
	$fIp.btnConnect configure -state normal
	"port1$p1SelIf" destroy
	"port2$p2SelIf" destroy

	close $fid
	dd destroy

	statusUpdate "Port monitor stopped..." black


		
}

proc baseWindowCreate {} {
	variable fTop
	variable lblStatus
	variable versionNumber
	cxLogDebug "[info level 0]"
	
	# Create the main window	
	wm withdraw .
	set top .top
	toplevel $top
	wm title $top "CxDiag"
	wm resizable $top 0 0
	# When the window is closed, kill the parent window too - closes the application
	wm protocol $top WM_DELETE_WINDOW [list destroy .]

	set topFrame $top.f
	frame $topFrame
	grid $topFrame - - - - -row 0 -column 0 -sticky ew
	set fTop $topFrame
	
	wm iconbitmap .top -default "./Images/Calnex_Icon.ico"

    
	set logo1 [image create photo -file Images/Logo_100_neo_100x30.png]
	label $topFrame.lblLogoNeo -image $logo1 -width 100 -height 30
	font create titleFont -family Ariel -size 14 -weight bold
	label $topFrame.lblTitle -text "Diagnostic Utility $versionNumber" -justify left -font titleFont
	set logo2 [image create photo -file Images/Calnex_rgb_trans_75x30.png]
	label $topFrame.lblLogoCalnex -image $logo2 -width 75 -height 30
	button $topFrame.bClose -text "Close" -width 8 -command {allDone}
	label $topFrame.lblFolderInfo -width 30 -wraplength 200 -anchor w
	button $topFrame.btnOpenFileExplorer -text "Open Data Folder" -command {openFileExplorer} -state disabled
	label $topFrame.lblStatusHeader -text "Status:"
	label $topFrame.lblStatus
	set lblStatus $topFrame.lblStatus
	
	grid $topFrame.lblLogoCalnex -row 0 -column 0 -sticky w -columnspan 2 -padx 5
	grid $topFrame.lblTitle -row 0 -column 4
	grid $topFrame.lblLogoNeo -row 0 -column 8 -sticky e -columnspan 1
	grid $topFrame.lblFolderInfo -row 1 -column 5 -sticky w -columnspan 3
	grid $topFrame.btnOpenFileExplorer -row 1 -column 8 -sticky w -padx 5
	grid $topFrame.lblStatusHeader -row 4 -column 0 -sticky e -pady 5 -padx 5
	grid $topFrame.lblStatus -row 4 -column 1 -sticky w -pady 5 -padx 5 -columnspan 6
	grid $topFrame.bClose -row 4 -column 8 -sticky e -pady 5 -padx 5

	return $topFrame
}

proc createIpLabelFrame {f ips} {
	variable fIp
	cxLogDebug "[info level 0]"
	
	set fIp $f.lfIp
	
	labelframe $f.lfIp -text "Instrument"
	label $f.lfIp.lblIp -text "Instrument IP Address: "
	label $f.lfIp.lblSnIntro -text "Model, Serial Number: "
	label $f.lfIp.lblSn
	#button $f.lfIp.btnConnect -text Connect -state disabled -command {connect $ipEntry .lfIp.lblSn .nbTabs.fSessions.lblMeasurementStatus .nbTabs.fSessions.lbSessionList .nbTabs.fSessions.bDownload .lfIp.cbIp}
	button $f.lfIp.btnConnect -text Connect -state disabled
	$f.lfIp.btnConnect configure -command "connect $f.lfIp.lblSn $f.lfIp.comboIp"
	#ttk::combobox $f.lfIp.cbIp -width 15 -textvariable ipEntry -validate all -validatecommand {cbIpAddress %P .lfIp.bConnect .lfIp.lblSn .nbTabs.fSessions.lblMeasurementStatus}
	ttk::combobox $f.lfIp.comboIp -width 15 -textvariable ipEntry
	$f.lfIp.comboIp configure -values $ips
	focus $f.lfIp.comboIp

	
	grid $f.lfIp -row 1 -column 0 -columnspan 5 -padx 10
	grid $f.lfIp.lblIp -row 1 -column 0 -sticky e
	#grid .eIp -row 0 -column 1 -sticky w
	grid $f.lfIp.comboIp -row 1 -column 1
	grid $f.lfIp.lblSnIntro -row 2 -column 0 -sticky e
	grid $f.lfIp.lblSn -row 2 -column 1 -sticky w -columnspan 2
	grid $f.lfIp.btnConnect -row 1 -column 2 -padx 5
	
	return $f.lfIp
}

proc sessionsTabCreate {tabsFrame} {
	variable btnGetSessionList
	variable fSessions
	variable fIp
	cxLogDebug "[info level 0]"
	
	set f [frame $tabsFrame.fSessions]
	set fSessions $f
	$tabsFrame add $f -text "Sessions"

	set infoText "This utility lists the session files on your instrument, allows you\
	to download the files and generate a pcap file from a PTP capture (if it exists in the session).\
	Only the latest 20 session files are shown.\
	It is not possible to download any session files if a measurement is running on your instrument.\
	Note that downloading the session files might take some time since they may be large."
	
	label $f.lblInfo -text $infoText -justify left -wraplength 500

	
	label $f.lblStatusIntro -text "Status: "
	label $f.lblStatus -width 50 -anchor w
	label $f.lblSessions -text "Last 20 Sessions: "
	#ttk::separator .sep -orient horizontal
	button $f.btnDownload -text "Download" -width 15 -state disabled -command "sessionsDownload $tabsFrame.fSessions.lbSessionList $tabsFrame.fSessions.lblStatus $fIp.btnConnect $tabsFrame.fSessions.btnDownload"
	button $f.btnGetSessionList -text "Get Session List" -width 15 -state disabled
	$f.btnGetSessionList configure -command "sessionsList $f.lbSessionList $f.lblStatus $f.btnDownload"
	set btnGetSessionList $f.btnGetSessionList
	listbox $f.lbSessionList -width 30 -height 20 -selectmode single
	checkbutton $f.cbGeneratePcap -text "Generate pcap" -variable sessionsDownloadGeneratePcapSelected
	
	grid $f.lblInfo -row 0 -column 0 -columnspan 10
	#grid $f.lblStatusIntro -row 0 -column 0 -sticky w
	#grid $f.lblStatus -row 0 -column 1 -sticky w -columnspan 6
	grid $f.lblSessions -row 1 -column 0 -sticky w -columnspan 2
	#grid $f.lbSessionList -row 2 -column 0 -columnspan 3 -padx 25 -pady 5 -rowspan 20
	grid $f.lbSessionList -row 2 -column 0 -columnspan 3 -padx 25 -pady 0 -rowspan 2 -sticky ne
	#grid .sep -row 1 -column 0
	#grid .bListSessions -row 1 -column 0 -sticky e
	grid $f.btnGetSessionList -row 1 -column 3 -sticky nw
	grid $f.btnDownload -row 2 -column 3 -sticky nw
	grid $f.cbGeneratePcap -row 2 -column 4 -sticky nw
}

proc reportsClear {} {
	variable fReports
	cxLogDebug "[info level 0]"
	
	$fReports.lbReportList delete 0 99
	$fReports.lblStatus configure -text ""
}

proc reportsTabCreate {tabsFrame} {
	variable btnGetReportList
	variable fReports
	variable fIp
	cxLogDebug "[info level 0]"
	
	set f [frame $tabsFrame.fReports]
	set fReports $f
	$tabsFrame add $f -text "Reports"

	set infoText "This utility lists the report files on your instrument and allows you\
	to download them.\
	Only the latest 20 report files are shown.\
	It is not possible to download any report files if a measurement is running on your instrument."
	
	label $f.lblInfo -text $infoText -justify left -wraplength 500

	
	label $f.lblStatusIntro -text "Status: "
	label $f.lblStatus -width 50 -anchor w
	label $f.lblReports -text "Last 20 Reports: "
	#ttk::separator .sep -orient horizontal
	button $f.btnDownload -text "Download" -width 15 -state disabled -command "reportsDownload $tabsFrame.fReports.lbReportList $tabsFrame.fReports.lblStatus $fIp.btnConnect $tabsFrame.fReports.btnDownload"
	button $f.btnGetReportList -text "Get Report List" -width 15 -state disabled
	$f.btnGetReportList configure -command "reportsList $f.lbReportList $f.lblStatus $f.btnDownload"
	set btnGetReportList $f.btnGetReportList
	listbox $f.lbReportList -width 30 -height 20 -selectmode single
	#checkbutton $f.cbGeneratePcap -text "Generate pcap" -variable sessionsDownloadGeneratePcapSelected
	
	grid $f.lblInfo -row 0 -column 0 -columnspan 10
	#grid $f.lblStatusIntro -row 0 -column 0 -sticky w
	#grid $f.lblStatus -row 0 -column 1 -sticky w -columnspan 6
	grid $f.lblReports -row 1 -column 0 -sticky w -columnspan 2
	#grid $f.lbSessionList -row 2 -column 0 -columnspan 3 -padx 25 -pady 5 -rowspan 20
	grid $f.lbReportList -row 2 -column 0 -columnspan 3 -padx 25 -pady 0 -rowspan 2 -sticky ne
	#grid .sep -row 1 -column 0
	#grid .bListSessions -row 1 -column 0 -sticky e
	grid $f.btnGetReportList -row 1 -column 3 -sticky nw
	grid $f.btnDownload -row 2 -column 3 -sticky nw
	#grid $f.cbGeneratePcap -row 2 -column 4 -sticky nw
}

proc reportsList {listboxReports lblStat btnDownload} {
	variable paragon
	variable lblStatus
	variable fReports
	variable fIp
	variable fTabs
	variable btnGetReportList

	cxLogDebug "[info level 0]"

	reportsClear
	$btnDownload configure -state disabled
	$btnGetReportList configure -state disabled
	$fIp.btnConnect configure -state disabled
	$fTabs state disabled

	set okToDownload TRUE

	statusUpdate "Getting list of reports..." "green"
	# It looks as though we can't tell whether any given session is associated
	# with a running measurement. So, we can't allow any download while
	# measurements are running.
	if {[paragon anyMeasurementRunning]} {
		statusUpdate "All measurement must be stopped before downloading" "red"
		set okToDownload FALSE
		$btnDownload configure -state disabled
	} else {
		statusUpdate "Getting list of reports... No measurements running" "green"
		#$btnDownload configure -state normal
	}
	#$lblStat configure -text $status -fg $statusFg
	set reps [paragon getAllReportFilenames]
	set reports [getLastReports $reps 20]

	$listboxReports delete 0 99
	if {[llength $reports] == 0} {
		$listboxReports insert 0 "No reports on instrument"
		set okToDownload FALSE
	} else {
		foreach report $reports {
			$listboxReports insert end $report	
		}
		$listboxReports selection set 0
	}
	$btnGetReportList configure -state normal
	$fIp.btnConnect configure -state normal
	if {$okToDownload} {
		$btnDownload configure -state normal
	} else {
		$btnDownload configure -state disabled
	}
	if {$okToDownload} {
#		statusUpdate "Getting list of sessions... Done" "black"
	}
	$fTabs state !disabled

}

proc reportsDownload {listboxReports measurementStatusLabel btnConnect btnDownload} {
	variable paragon
	variable downloadedDataFolderPrefix
	variable cxDiagMode
	
	cxLogDebug "[info level 0]"
	set index [$listboxReports curselection]
	set report [$listboxReports get $index]
	#puts "download $index $folder"
	
	$measurementStatusLabel configure -text "Download in progress..." -fg "green"
	statusUpdate "Download in progress..." "green"
	$btnConnect configure -state disabled
	$btnDownload configure -state disabled
	#CxDebugData create dd paragon $downloadedDataFolderPrefix $cxDiagMode
	#dd writeDebugInfoFile
	#dd downloadData $folder
	#dd destroy

	set sn [paragon getSerialNumber]
	set folderName [file join "[pwd]" "$downloadedDataFolderPrefix$sn"]
	set root $folderName
	set destReportFolder "$root/Reports"
	file mkdir $destReportFolder


    paragon downloadReportFile $report $destReportFolder

	$measurementStatusLabel configure -text "Download complete" -fg "black"
	statusUpdate "Download in progress... done" "black"
	$btnConnect configure -state normal
	$btnDownload configure -state normal

}

proc debugInfoPortsCreate {parent} {	
	cxLogDebug "[info level 0]"
	set f ${parent}
	#puts $f

	#labelframe $f -text "Port Details"
	ttk::treeview $f.tvPortDetails -columns "Port1 Port2" -displaycolumns "Port1 Port2" -height 7 -selectmode none
	$f.tvPortDetails column #0 -width 150 -anchor e
	$f.tvPortDetails column Port1 -width 150 -anchor w
	$f.tvPortDetails column Port2 -width 150 -anchor w
	$f.tvPortDetails heading Port1 -text "Port 1" -anchor w
	$f.tvPortDetails heading Port2 -text "Port 2" -anchor w
	grid $f.tvPortDetails
	#.tree insert {} end -id Languages -text "Languages"
	#.tree insert Languages end -text C -values [list "Dennis Ritchie" "1990"]
	#.tree insert "" end -id Direct -text "Direct"
	
	# Hard-coded attributes of a port (can't query this from the instrument
	# since this information is not available if a port is not selected)
	set keys {"InterfaceType" "Fitted" "LineRate" "Fec" "Reach" "Manufacturer" "SerialNumber"}
	#puts $keys
	# Populate the table headings
	foreach key $keys {
		$f.tvPortDetails insert {} end -id $key -text $key
		$f.tvPortDetails item $key -tags [list -background red]
		#ttk::style configure Treeview -fieldbackground red
	
	}
	# Colour alternating lines of the listbox
	for {set i 0} {$i < [llength $keys]} {incr i 2} {
	    #$f.lbPortDetailsTitles itemconfigure $i -background #f0f0ff		
		#$f.tvPortDetails item $i -background #f0f0ff
	}

		
	#label $f.lblSelectedIfIntro -text "Selected Interface:"
	#label $f.lblSelectedIfLineRateIntro -text "Line Rate:"
	#label $f.lblSelectedIfFecIntro -text "FEC:"
	#label $f.lblSelectedIfReachIntro -text "Reach:"
	#label $f.lblSelectedIfManufacturerIntro -text "Manufacturer:"
	#label $f.lblSelectedIfPartNumberIntro -text "Model:"
	#label $f.lblSelectedIfSerialNumberIntro -text "Serial Number:"
	#label $f.lblSelectedIf -width 40
	#label $f.lblSelectedIfLineRate
	#label $f.lblSelectedIfFec
	#label $f.lblSelectedIfReach
	#label $f.lblSelectedIfManufacturer
	#label $f.lblSelectedIfPartNumber
	#label $f.lblSelectedIfSerialNumber
	
	#grid $f -row 5 -column 1 -columnspan 5 -padx 10
	#grid $f.lblSelectedIfIntro -row 0 -column 0 -sticky e
	#grid $f.lblSelectedIfLineRateIntro -row 1 -column 0 -sticky e
	#grid $f.lblSelectedIfFecIntro -row 2 -column 0 -sticky e
	#grid $f.lblSelectedIfReachIntro -row 3 -column 0 -sticky e
	#grid $f.lblSelectedIfManufacturerIntro -row 4 -column 0 -sticky e
	#grid $f.lblSelectedIfPartNumberIntro -row 5 -column 0 -sticky e
	#grid $f.lblSelectedIfSerialNumberIntro -row 6 -column 0 -sticky e
	#
	#grid $f.lblSelectedIf -row 0 -column 1 -sticky w
	#grid $f.lblSelectedIfLineRate -row 1 -column 1 -sticky w
	#grid $f.lblSelectedIfFec -row 2 -column 1 -sticky w
	#grid $f.lblSelectedIfReach -row 3 -column 1 -sticky w
	#grid $f.lblSelectedIfManufacturer -row 4 -column 1 -sticky w
	#grid $f.lblSelectedIfPartNumber -row 5 -column 1 -sticky w
	#grid $f.lblSelectedIfSerialNumber -row 6 -column 1 -sticky w


}

proc debugInfoTabCreate {tabsFrame} {
	variable fDebugInfo
	cxLogDebug "[info level 0]"

	set f [frame $tabsFrame.fDebugInfo]
	set fDebugInfo $f
	$tabsFrame add $f -text "Debug Info"
	
	# Instrument Details
	button $f.btnGetDebugInfo -text "Read from Instrument" -width 20 -state disabled -command "debugInfoGet $f.lfInstrDetails.lblSwVersion $f.lfInstrDetails.lblOsVersion $f.lfInstrDetails.lblHwCapabilities $f.tvPortDetails $f.btnDownloadDebugInfo"
	button $f.btnDownloadDebugInfo -text "Download" -width 20 -state disabled -command "debugInfoDownload"
	labelframe $f.lfInstrDetails -text "Instrument Details"
	label $f.lfInstrDetails.lblSwVersionIntro -text "Software Version:"
	label $f.lfInstrDetails.lblOsVersionIntro -text "OS Version:"
	label $f.lfInstrDetails.lblHwCapabilitiesIntro -text "HW Capabilities:"
	label $f.lfInstrDetails.lblSwVersion -width 30 -anchor w
	label $f.lfInstrDetails.lblOsVersion -width 30 -anchor w
	label $f.lfInstrDetails.lblHwCapabilities -width 30 -anchor w

	grid $f.lfInstrDetails  -row 0 -column 0 -columnspan 2 -rowspan 5 -padx 10 -sticky w
	grid $f.btnGetDebugInfo -row 0 -column 2 -padx 10 -sticky w
	grid $f.btnDownloadDebugInfo -row 1 -column 2 -padx 10 -sticky w
	grid $f.lfInstrDetails.lblSwVersionIntro -row 0 -column 0 -sticky e
	grid $f.lfInstrDetails.lblOsVersionIntro -row 1 -column 0 -sticky e
	#grid $f.lfInstrDetails.lblHwCapabilitiesIntro -row 2 -column 0 -sticky e
	grid $f.lfInstrDetails.lblSwVersion -row 0 -column 1 -sticky w
	grid $f.lfInstrDetails.lblOsVersion -row 1 -column 1 -sticky w
	#grid $f.lfInstrDetails.lblHwCapabilities  -row 2 -column 1 -sticky w

	#createDebugInfoPorts "lfPortDetails" $f
	#debugInfoPortsCreate $f
	opticalModuleInfoPortsCreate $f
	grid $f.tvPortDetails -row 5 -column 0 -columnspan 5 -padx 10 -sticky w
	ttk::style configure Treeview -background gray94
	ttk::style configure Treeview.Heading -background gray94
	#createDebugInfoPort "lfPort2Details" $f
	#grid $f.lfPort2Details -row 10 -column 1 -columnspan 5 -padx 10 -sticky w


}

proc allPacketCaptureTabCreate {tabsFrame} {
	variable fAllPacketCapture
	variable cxDiagMode
	cxLogDebug "[info level 0]"

	set f [frame $tabsFrame.fAllPacketCapture]
	set fAllPacketCapture $f
	if {$cxDiagMode != "BASIC"} {
	 	$tabsFrame add $f -text "All Packet Capture"
	}

	set infoText "This utility configures Paragon to capture all packets present on the link on port 2.\
	Before starting an all packet capture, the link on Port 2 must established and no\
	measurements should be running."
	
	label $f.lblInfo -text $infoText -justify left -wraplength 500
	label $f.lblPort2LinkHeader -text "Port 2 Link Status:"
	label $f.lblPort2Link
	label $f.lblStatusHeader -text "Status:"
	label $f.lblStatus
	button $f.btnRefreshStatus -text "Refresh Status" -width 12 -state normal -command "allPacketCaptureRefreshStatus"
	label $f.lblCaptureDurationIntro -text "Capture Duration (s):"
	entry $f.entryCaptureDuration
	$f.entryCaptureDuration insert 0 "30"
	button $f.btnStartCapture -text "Start Capture" -width 12 -state disabled -command "allPacketCaptureStart $f"
	label $f.lblCaptureSessionNameHeader -text "Session:"
	label $f.lblCaptureSessionName -width 25
	button $f.btnPfv -text "View in PFV" -width 12 -state disabled -command "launchPfv $f"
	button $f.btnDownload -text "Download" -width 12 -state disabled -command "allPacketCaptureDownloadSession $f"
	
	grid $f.lblInfo -row 0 -column 0 -columnspan 10
	grid $f.lblPort2LinkHeader -row 1 -column 0 -sticky e
	grid $f.lblPort2Link -row 1 -column 1 -sticky w
	grid $f.lblStatusHeader -row 2 -column 0 -sticky e
	grid $f.lblStatus -row 2 -column 1 -sticky w -columnspan 10
	grid $f.btnRefreshStatus -row 1 -column 4 -columnspan 2 -rowspan 2 -sticky w
	grid $f.lblCaptureDurationIntro -row 3 -column 0 -sticky e
	grid $f.entryCaptureDuration -row 3 -column 1 -sticky w 
	grid $f.btnStartCapture -row 3 -column 4 -sticky w -columnspan 2
	grid $f.lblCaptureSessionNameHeader -row 4 -column 0 -sticky e
	grid $f.lblCaptureSessionName -row 4 -column 1 -sticky w -columnspan 2
	grid $f.btnPfv -row 4 -column 4 -sticky w
	grid $f.btnDownload -row 4 -column 5 -sticky w
}

proc opticalModuleInfoPortsCreate {parent} {	
	set f ${parent}
	cxLogDebug "[info level 0]"
	#puts $f

	ttk::treeview $f.tvPortDetails -columns "Port1 Port2" -displaycolumns "Port1 Port2" -height 13 -selectmode none
	$f.tvPortDetails column #0 -width 150 -anchor e
	$f.tvPortDetails column Port1 -width 150 -anchor w
	$f.tvPortDetails column Port2 -width 150 -anchor w
	$f.tvPortDetails heading Port1 -text "Port 1" -anchor w
	$f.tvPortDetails heading Port2 -text "Port 2" -anchor w
	grid $f.tvPortDetails
	
}

proc opticalModuleEepromLabelFrameCreate {parent} {
	variable fOpticalModuleEepromLabelFrame
	cxLogDebug "[info level 0]"
	
	set fOpticalModuleEepromLabelFrame $parent.lfEepromDetails
	set f $fOpticalModuleEepromLabelFrame
	label $f.lblTypeHeader -text "Type:"
	label $f.lblType -width 30
	label $f.lblEncodingHeader -text "Encoding:"
	label $f.lblEncoding -width 30
	label $f.lblBrNominalHeader -text "BR Nominal:"
	label $f.lblBrNominal -width 30
	label $f.lblExtComplianceHeader -text "Extended Compliance:"
	label $f.lblExtCompliance -width 30
	button $f.btnGetSelectedEepromContents -text "Read" -state disabled -command opticalModuleSelectedEepromContentsRead
	button $f.btnStartStopPowerMeasurement -text "Start Reading Power Levels" -width 10 -state disabled -command opticalModuleStartPowerMeasurement
	ttk::treeview $f.tvPowerDetails -columns "Rx Tx" -displaycolumns "Rx Tx" -height 4 -selectmode none
	$f.tvPowerDetails column #0 -width 150 -anchor e
	$f.tvPowerDetails column Rx -width 150 -anchor w
	$f.tvPowerDetails column Tx -width 150 -anchor w
	$f.tvPowerDetails heading Rx -text "Rx" -anchor w
	$f.tvPowerDetails heading Tx -text "Tx" -anchor w


	#grid $f.lfIp -row 1 -column 0 -columnspan 5 -padx 10
	grid $f.btnGetSelectedEepromContents -row 0 -column 3 -padx 10 -sticky ne
	grid $f.lblTypeHeader -row 0 -column 0 -sticky e
	grid $f.lblType -row 0 -column 1 -sticky w
	grid $f.lblEncodingHeader -row 1 -column 0 -sticky e
	grid $f.lblEncoding -row 1 -column 1 -sticky w
	grid $f.lblBrNominalHeader -row 2 -column 0 -sticky e
	grid $f.lblBrNominal -row 2 -column 1 -sticky w
	grid $f.lblExtComplianceHeader -row 3 -column 0 -sticky e
	grid $f.lblExtCompliance -row 3 -column 1 -sticky w
	#grid $f.tvPowerDetails -columnspan 3

	# Commented out until I figure out how to do it.
	#grid $f.btnStartStopPowerMeasurement -row 5 -column 2 -sticky se
	
	return $f
}

proc opticalModuleEepromWriteLabelFrameCreate {parent} {
	variable fOpticalModuleEepromWriteLabelFrame
	cxLogDebug "[info level 0]"
	
	set fOpticalModuleEepromWriteLabelFrame $parent.lfEepromWrite
	set f $fOpticalModuleEepromWriteLabelFrame
	label $f.lblWarning -text "Warning: Use at your own risk. You may make your module unusable." -justify left
	label $f.lblPageHeader -text "Page"
	label $f.lblAddrHeader -text "Addr\n(dec)"
	label $f.lblDataHeader -text "Data\n(hex)"
	ttk::combobox $f.comboPage -width 7 -textvariable eepromPageSelect
	entry $f.entryAddr -width 7
	entry $f.entryData -width 7
	button $f.btnOpticalModuleSelectedEepromAddressRead -text "Read" -width 6 -state disabled -command opticalModuleSelectedEepromAddressRead
	button $f.btnOpticalModuleSelectedEepromAddressWrite -text "Write" -width 6 -state disabled -command opticalModuleSelectedEepromAddressWrite
	label $f.lblMemoryMap -text "For SPF28 and SFPPLUS, refer to SFF-8472\nFor QSFP28 and QSFPPLPUS, refer to SFF-8636" -justify left

	#grid $f.lfIp -row 1 -column 0 -columnspan 5 -padx 10
	grid $f.lblWarning -row 0 -column 0 -sticky w -columnspan 10
	grid $f.lblPageHeader -row 1 -column 0 -sticky w
	grid $f.lblAddrHeader -row 1 -column 1 -sticky w
	grid $f.lblDataHeader -row 1 -column 2 -sticky w
	grid $f.comboPage -row 2 -column 0 -sticky w -padx 1
	grid $f.entryAddr -row 2 -column 1 -sticky w 
	grid $f.entryData -row 2 -column 2 -sticky w 
	grid $f.btnOpticalModuleSelectedEepromAddressRead -row 2 -column 3 -sticky w
	grid $f.btnOpticalModuleSelectedEepromAddressWrite -row 2 -column 4 -sticky w
	grid $f.lblMemoryMap -row 3 -column 0 -sticky w  -columnspan 6
	#grid $f.tvPowerDetails -columnspan 3

	# Commented out until I figure out how to do it.
	#grid $f.btnStartStopPowerMeasurement -row 5 -column 2 -sticky se
	
	return $f
}

proc opticalModuleInfoTabCreate {tabsFrame} {
	variable fOpticalModuleInfo
	variable cxDiagMode
	cxLogDebug "[info level 0]"
	
	set f [frame $tabsFrame.fOpticalModuleInfo]
	set fOpticalModuleInfo $f
	if {$cxDiagMode != "BASIC"} {
		$tabsFrame add $f -text "OpticalModuleInfo"
	}
	
	set infoText "This utility reads the registers from an optical module in a specified port.\
	The list of interfaces presented below are those that are available on your instrument.\
	You must make sure that you select the interface / module type that you have inserted - this utility\
	is intended to interrogate the register contents even if the instrument does not recognise the module i.e.\
	even if the module shows up as \"Not Fitted\" in the UI.\n\
	Note: Downloading the data might take some time since the instrument logs may be large."
	
	label $f.lblInfo -text $infoText -justify left -wraplength 500

	button $f.btnGetOpticalModuleInterfaces -text "Get Interfaces" -width 20 -state disabled -command "opticalModuleInfoGet"
	button $f.btnDownloadOpticalModuleInfo -text "Download" -width 20 -state disabled -command "opticalModuleInfoDownload"
	label $f.lblStatusHeader -text "Status:"
	label $f.lblStatus

	#opticalModuleInfoPortsCreate $f
	#grid $f.tvPortDetails -row 5 -column 0 -columnspan 5 -padx 10 -sticky w
	#ttk::style configure Treeview -background gray94
	#ttk::style configure Treeview.Heading -background gray94

	label $f.lblPortSelectHeader -text "Select Port:"
	ttk::combobox $f.comboPortSelect -width 15 -textvariable portSelect
	$f.comboPortSelect configure -values {"Port 1" "Port 2"}
	$f.comboPortSelect current 0
	label $f.lblIfSelectHeader -text "Select Interface:"
	ttk::combobox $f.comboIfSelect -width 15 -textvariable ifSelect

	label $f.lblGpioStatusHeader -text "GPIO Status:"
	label $f.lblGpioStatus
	label $f.lblSelectedInterfaceHeader -text "Active Interface:"
	label $f.lblSelectedInterface
	labelframe $f.lfEepromDetails -text "Selected Register Contents"
	opticalModuleEepromLabelFrameCreate $f
	
	grid $f.lblInfo -row 0 -column 0 -columnspan 10
	#grid $f.lblStatusHeader -row 1 -column 0 -sticky e
	#grid $f.lblStatus -row 1 -column 1 -sticky w -columnspan 8
	grid $f.lblPortSelectHeader -row 2 -column 0 -sticky e
	grid $f.comboPortSelect -row 2 -column 1 -sticky w
	grid $f.btnGetOpticalModuleInterfaces -row 2 -column 2 -padx 10 -sticky w
	grid $f.lblIfSelectHeader -row 3 -column 0 -sticky e
	grid $f.comboIfSelect -row 3 -column 1 -sticky w
	grid $f.btnDownloadOpticalModuleInfo -row 3 -column 2 -padx 10 -sticky w
	grid $f.lblGpioStatusHeader -row 4 -column 0 -sticky e
	grid $f.lblGpioStatus -row 4 -column 1 -sticky w -columnspan 8
	grid $f.lblSelectedInterfaceHeader -row 5 -column 0 -sticky e
	grid $f.lblSelectedInterface -row 5 -column 1 -sticky w -columnspan 8
	grid $f.lfEepromDetails -row 6 -column 0 -columnspan 8 -sticky w

	bind $f.comboIfSelect <<ComboboxSelected>> comboIfSelectionChanged
	bind $f.comboPortSelect <<ComboboxSelected>> comboPortSelectionChanged
	
	# In advanced mode we add the ability to write to the EEPROM in the module
	labelframe $f.lfEepromWrite -text "Register Write"
	opticalModuleEepromWriteLabelFrameCreate $f
	if {$cxDiagMode == "ADVANCED"} {
		grid $f.lfEepromWrite -row 8 -column 0 -columnspan 8 -sticky w
	}
}

proc portMonitorItemSelectionLabelFrameCreate {parent} {
	variable fPortMonitorItemSelectionLabelFrame
	cxLogDebug "[info level 0]"
	
	set fPortMonitorItemSelectionLabelFrame $parent.lfItemSelection
	set f $fPortMonitorItemSelectionLabelFrame

	checkbutton $f.cbRxPower -text "Rx Power (mW)" -variable portMonitorRxPowerSelected
	checkbutton $f.cbRxPowerAlarms -text "Rx Power Alarms (hex)" -variable portMonitorRxPowerAlarmsSelected
	checkbutton $f.cbTxPower -text "Tx Power (mW)" -variable portMonitorTxPowerSelected
	checkbutton $f.cbLos -text "LOS (hex)" -variable portMonitorLosSelected
	checkbutton $f.cbBips -text "BIPS (hex)" -variable portMonitorBipsSelected
	checkbutton $f.cbFecCorr -text "FEC corrected (hex)" -variable portMonitorFecCorrSelected
	checkbutton $f.cbFecUncorr -text "FEC uncorrected (hex)" -variable portMonitorFecUncorrSelected
	checkbutton $f.cbFecLock -text "FEC lock status (hex)" -variable portMonitorFecLockSelected
	
	grid $f.cbRxPower -row 0 -column 0 -sticky w
	grid $f.cbTxPower -row 0 -column 1 -sticky w
	grid $f.cbRxPowerAlarms -row 0 -column 2 -sticky w
	grid $f.cbLos -row 0 -column 3 -sticky w
	grid $f.cbFecCorr -row 1 -column 0 -sticky w
	grid $f.cbFecUncorr -row 1 -column 1 -sticky w
	grid $f.cbFecLock -row 1 -column 2 -sticky w
	grid $f.cbBips -row 1 -column 3 -sticky w
}

proc portMonitorTabCreate {tabsFrame} {
	variable fPortMonitor
	variable cxDiagMode
	cxLogDebug "[info level 0]"
	
	set f [frame $tabsFrame.fPortMonitor]
	set fPortMonitor $f
	if {$cxDiagMode != "BASIC"} {
		$tabsFrame add $f -text "PortMonitor"
	}
	
	set infoText "This utility monitors specific activity associated with the \
					currently selected interface on each instrument port."
	
	label $f.lblInfo -text $infoText -justify left -wraplength 500

	button $f.btnRefreshStatus -text "Refresh Status" -width 12 -state normal -command "portMonitorRefreshStatus"
	button $f.btnStartStopMonitoring -text "Start" -width 12 -state disabled -command "portMonitorStart"
	label $f.lblPort1SelectedInterfaceHeader -text "Port 1 Interface:"
	label $f.lblPort1SelectedInterface -justify left
	label $f.lblPort1SelectedInterfaceFecHeader -text "FEC:"
	label $f.lblPort1SelectedInterfaceFec -justify left
	label $f.lblPort2SelectedInterfaceHeader -text "Port 2 Interface:"
	label $f.lblPort2SelectedInterface -justify left
	label $f.lblPort2SelectedInterfaceFecHeader -text "FEC:"
	label $f.lblPort2SelectedInterfaceFec -justify left
	labelframe $f.lfItemSelection -text "Select Items to Monitor"
	ttk::treeview $f.tvMonitorData -columns "Port1 Port2" -displaycolumns "Port1 Port2" -height 8 -selectmode none
	$f.tvMonitorData column #0 -width 110 -anchor e
	$f.tvMonitorData column Port1 -width 150 -anchor w
	$f.tvMonitorData column Port2 -width 150 -anchor w
	$f.tvMonitorData heading Port1 -text "Port 1" -anchor w
	$f.tvMonitorData heading Port2 -text "Port 2" -anchor w
	
	grid $f.lblInfo -row 0 -column 0 -columnspan 8 -sticky w
	grid $f.btnRefreshStatus -row 1 -column 6 -padx 10 -sticky w
	grid $f.btnStartStopMonitoring -row 2 -column 6 -padx 10 -sticky w
	grid $f.lblPort1SelectedInterfaceHeader -row 1 -column 0 -sticky e
	grid $f.lblPort1SelectedInterface -row 1 -column 1 -sticky w 
	grid $f.lblPort1SelectedInterfaceFecHeader -row 1 -column 2 -sticky e
	grid $f.lblPort1SelectedInterfaceFec -row 1 -column 3 -sticky w 
	grid $f.lblPort2SelectedInterfaceHeader -row 2 -column 0 -sticky e
	grid $f.lblPort2SelectedInterface -row 2 -column 1 -sticky w 
	grid $f.lblPort2SelectedInterfaceFecHeader -row 2 -column 2 -sticky e
	grid $f.lblPort2SelectedInterfaceFec -row 2 -column 3 -sticky w 
	grid $f.lfItemSelection -row 3 -column 0 -columnspan 8
	grid $f.tvMonitorData -row 4 -column 0 -columnspan 8

	portMonitorItemSelectionLabelFrameCreate $f
}

proc tabsCreate {f} {
	#variable ipLabelFrame
	variable fTabs
	cxLogDebug "[info level 0]"
	
	ttk::notebook $f.nbTabs -width 500
	set tabsFrame $f.nbTabs
	set fTabs $tabsFrame
	
	sessionsTabCreate $tabsFrame
	reportsTabCreate $tabsFrame
	debugInfoTabCreate $tabsFrame
	allPacketCaptureTabCreate $tabsFrame
	opticalModuleInfoTabCreate $tabsFrame
	portMonitorTabCreate $tabsFrame
	bind $f.nbTabs <<NotebookTabChanged>> nbTabsTabChanged
	grid $f.nbTabs -row 3 -column 0 -columnspan 10 -sticky w -padx 10 -pady 10
	return $f.nbTabs
}

proc configureUI {ipAddresses} {
	#variable topFrame
	#variable ipLabelFrame
	#variable tabsFrame
	cxLogDebug "[info level 0]"
	
	set topFrame [baseWindowCreate]
	set ipLabelFrame [createIpLabelFrame $topFrame $ipAddresses]
	set tabsFrame [tabsCreate $topFrame]
	
	$ipLabelFrame.comboIp configure -validate all -validatecommand "cbIpAddress %P .top.f.lfIp.btnConnect .top.f.lfIp.lblSn .top.f.nbTabs.fSessions.lblMeasurementStatus"
	
}

proc getLastUsedIpAddresses {} {
	variable lastUsedIpAddressListFilename
	cxLogDebug "[info level 0]"
	set data {}
	set f $lastUsedIpAddressListFilename
	if {[file exists $f]} {
		set fid [open $f "r"]
	    set data [split [string trim [read $fid]]]
	    close $fid
	}
	#puts $data
    return $data
}
proc addIpAddressToLastUsedList {addr} {
	variable lastUsedIpAddressListFilename
	cxLogDebug "[info level 0]"
	set currentList [getLastUsedIpAddresses]
	if {$addr in $currentList} {
		# Do nothing
	} else {
		lappend currentList $addr
		set f $lastUsedIpAddressListFilename
		set fid [open $f "w"]
		foreach a $currentList {
			puts $fid $a
		}
		close $fid	
	}
}

proc getCxDiagMode {} {
	global cxDiagMode
	cxLogDebug "[info level 0]"
	return $cxDiagMode
}

package require Tk
package require tooltip


set lastUsedIpAddressListFilename "./lastUsedIpAddresses.txt"

set lastUsedIpAddresses [getLastUsedIpAddresses]
# If readRegister methods don't exist then use basic mode
set cxDiagMode "BASIC"
if {[info commands CxRegister] == ""} {
	set cxDiagMode "BASIC"
} else {
	if {[info commands CxRegisterOpticalModuleWrite] == ""} {
		set cxDiagMode "FULL"
	} else {
		set cxDiagMode "ADVANCED"
	}
	cxLogDebug "cxDiagMode: $cxDiagMode"
}
configureUI $lastUsedIpAddresses


