################################################################################
# CxInstrumentOpticalModuleClass.tcl
# ------------------------------------------------------------------------------
# This script is provided as a utility supporting Paragon-neo and Paragon-100G.
# It is NOT intended as example code for controlling these instruments.
# No part of this script should be used independently of this script without
# the explicit written permission of Calnex Solutions Ltd.
################################################################################
source calnexRest.tcl
source CxInstrumentOpticalModuleEepromClass.tcl

	
oo::class create CxInstrumentOpticalModule {
	variable Type
	variable NumberOfChannels
	variable PortName
	variable TypeForRest
	#variable NumberOfRegisters
	variable A0device
	variable A2device
	variable gpioSettings
	#variable eepromDict
	#variable eepromFields
	variable eeprom
	variable moduleDict
	#variable Banks
	variable DeviceAddresses
	variable ModuleRegisters
	variable FpgaRegisters
	#variable Instrument
	#variable IpAddress
	variable RegisterReaderFileHandle
	variable RegisterReaderPage
	variable CurrentPageSelect
	# Device numbering is 6=U41, 7=U37, 8=U40
	
	constructor {portName transceiverType} {
		cxLogDebug "[info level 0]"
		set Type [string toupper $transceiverType]
		set PortName $portName
		#set Instrument $instrument
		set RegisterReaderFileHandle ""
		
		CxInstrumentOpticalModuleEeprom create eeprom $Type

		dict set gpioDict "SFPPLUS" { \
			"Port 1" {device 7 bank 4 pin 3} \
			"Port 2" {device 8 bank 4 pin 3}}
		dict set gpioDict "SFP28" 	{ \
			"Port 1" {device 7 bank 4 pin 3} \
			"Port 2" {device 8 bank 4 pin 3}}
		dict set gpioDict "CXP" 	{ \
			"Port 1" {device 7 bank 0 pin 3} \
			"Port 2" {device 8 bank 0 pin 3}}
		dict set gpioDict "QSFP28" 	{ \
			"Port 1" {device 7 bank 2 pin 3} \
			"Port 2" {device 8 bank 2 pin 3}}
		dict set gpioDict "QSFPPLUS" 	{ \
			"Port 1" {device 7 bank 2 pin 3} \
			"Port 2" {device 8 bank 2 pin 3}}
		dict set gpioDict "SFP1G" 	{ \
			"Port 1" {device 7 bank 3 pin 3} \
			"Port 2" {device 8 bank 3 pin 3}}
		dict set gpioDict "SFP100M" 	{ \
			"Port 1" {device 6 bank 3 pin 3} \
			"Port 2" {device 6 bank 4 pin 3}}
		dict set gpioDict "CFP4" 	{ \
			"Port 1" {device 7 bank 1 pin 3} \
			"Port 2" {device 8 bank 1 pin 3}}
		# Not sure about CFP2 - dummy entry for now
		dict set gpioDict "CFP2" 	{ \
			"Port 1" {device 0 bank 0 pin 0} \
			"Port 2" {device 0 bank 0 pin 0}}
		dict set gpioDict "NONE" 	{ \
			"Port 1" {device 0 bank 0 pin 0} \
			"Port 2" {device 0 bank 0 pin 0}}


		# The mapping of modules to the parameters for REST is a little confusing.
		# Interface		Neo Debug Device Type	P100G Debug Device Type		Comments
		# SFP100M		SFP						NA							100M optical
		# SFP1G			SFPPLUS					NA							1G optical
		# SFP+			SFP28					SFP28						10G optical
		# SFP28			SFP28					SFP28						25G optical
		# QSFP+			QSFP28					QSFP28						40G optical
		# QSFP28		QSFP28					QSFP28						100G optical
		# CXP			CXP						CXP							100G optical
		# CFP4			CFP4					CFP4						100G optical
		#
		# Addresses:
		#	SFP28/QSFP28 (port 1) - BIP: 0x2820A4; FECcorr: 0x436C; FECuncorr: 0x4368; FEClock: 0x4364
		#	SFP28/QSFP28 (port 2) - BIP: 0x2822A4; FECcorr: 0x437C; FECuncorr: 0x4378; FEClock: 0x4374
		#	SFPPLUS (port 1) - BIP: 0x2820A4
		#	SFPPLUS (port 2) - BIP: 0x2822A4
		#	QSFPPLUS (port 1) - BIP: 0x28210C
		#	QSFPPLUS (port 2) - BIP: 0x28230C
		#	CFP4 (port 1) - FECcorr: 0x436C; FECuncorr: 0x4368; FEClock: 0x4364
		#	CFP4 (port 2) - FECcorr: 0x437C; FECuncorr: 0x4378; FEClock: 0x4374
		dict set moduleDict "SFP28" 		{typeForRest SFP28 		numberOfChannels 1 addresses {"Port 1" {A0 1 A2 0 BIP 0 BIPaddr 2629796 FECcorr 17260 FECuncorr 17256 FEClock 17252} 	"Port 2" {A0 3 A2 2 BIP 0 BIPaddr 2630308 FECcorr 17276 FECuncorr 17272 FEClock 17268}}}
		dict set moduleDict "SFPPLUS" 		{typeForRest SFP28 		numberOfChannels 1 addresses {"Port 1" {A0 1 A2 0 BIP 0 BIPaddr 2629796 FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 3 A2 2 BIP 0 BIPaddr 2630308 FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "QSFP28" 		{typeForRest QSFP28 	numberOfChannels 4 addresses {"Port 1" {"00" 0 "01" 0 "02" 0 "03" 0 "20" 0 "21" 0 BIP 0 BIPaddr 2629796 FECcorr 17260 FECuncorr 17256 FEClock 17252} 		"Port 2" {"00" 1 "01" 1 "02" 1 "03" 1 "20" 1 "21" 1 BIP 0 BIPaddr 2630308 FECcorr 17276 FECuncorr 17272 FEClock 17268}}}
		dict set moduleDict "QSFPPLUS" 		{typeForRest QSFP28 	numberOfChannels 4 addresses {"Port 1" {"00" 0 "01" 0 "02" 0 "03" 0 "20" 0 "21" 0 BIP 0 BIPaddr 2629900 FECcorr NA FECuncorr NA FEClock NA} 				"Port 2" {"00" 1 "01" 1 "02" 1 "03" 1 "20" 1 "21" 1 BIP 0 BIPaddr 2630412 FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "CXP" 			{typeForRest CXP 		numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "CFP4" 			{typeForRest CFP4 		numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr 17260 FECuncorr 17256 FEClock 17252} 	"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr 17276 FECuncorr 17272 FEClock 17268}}}
		dict set moduleDict "CFP2" 			{typeForRest CFP2 		numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "RJ45_1G"		{typeForRest RJ45_1G 	numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "RJ45_LOWRATE"	{typeForRest RJ45_LOWRATE 	numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "SFP1G"			{typeForRest SFPPLUS	numberOfChannels 0 addresses {"Port 1" {A0 1 A2 0 BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "SFP100M"		{typeForRest SFP	 	numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}
		dict set moduleDict "NONE"			{typeForRest NONE 		numberOfChannels 0 addresses {"Port 1" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA} 			"Port 2" {A0 NA A2 NA BIP NA BIPaddr NA FECcorr NA FECuncorr NA FEClock NA}}}

		set TypeForRest [dict get $moduleDict $Type typeForRest]
		#set Banks [dict get $moduleDict $Type banks]
		set DeviceAddresses [dict get $moduleDict $Type addresses $PortName]
		set NumberOfChannels [dict get $moduleDict $Type numberOfChannels]
		set gpio [dict get $gpioDict $Type]
		set gpioSettings [dict get $gpio $PortName]

		if {[getCxDiagMode] == "ADVANCED"} {
			CxRegisterOpticalModuleWrite create ModuleRegistersWrite $TypeForRest
		}
		CxRegister create ModuleRegisters "IF_JIT" $TypeForRest		
		CxRegister create FpgaRegisters "IF_JIT" "FPGA"	
		CxRegister create GpioRegisters "IF_JIT" "DIGIO"	
		#if {$Type == "SFP28" || $Type == "SFPPLUS"} {
		#	#set NumberOfRegisters 70
		#	set PortType "SFP28"
		#	set gpio [dict get $gpioDict $Type]
		#	set gpioSettings [dict get $gpio $PortName]
		#} 
		#if {$Type == "QSFP28" || $Type == "QSFPPLUS"} {
			#set NumberOfRegisters 70
		#	set PortType "QSFP28"
		#	set gpio [dict get $gpioDict $Type]
		#	set gpioSettings [dict get $gpio $PortName]
		#} 
		#if {$PortName == "Port1"} {
		#	set A0device 1
		#	set A2device 0
		#} else {
		#	set A0device 3
		#	set A2device 2			
		#}
		
		# Some pages will be valid, some won't. The only way to tell is to
		# write to the page select register; if the register does not reset to zero,
		# then the page is valid. However, we don't want to do this every time
		# this object is instantiated - takes too long. So, don't do this...
		set CurrentPageSelect "00"
		if {$Type == "QSFP28" || $Type == "QSFPPLUS"} {
			#my WritePageSelect "00"
			#my writeRegister "00" 127 "00"
			# Force the page select. Don't use writeRegister since this messes with the paging.
			set deviceAddress [dict get $DeviceAddresses "00"]
			set regVal [ModuleRegisters registerSet2 "00" $deviceAddress 127 0]
		}
	}
	
	destructor {
		cxLogDebug "[info level 0]"
		#puts "Destroying optical module: $PortName $Type"
	}
	
	# These are right. Spec says "only required if" which doesn't mean
	# that they aren't implemented....
#	method SetValidPages {} {
		# Page 03h is only required if page 00h byte 2 bit 2 is low
#		set byteDec [lindex [my readRegister "00" 2] 0]
#		if {$byteDec & 1 == 1} {
#			eeprom removePage "03"
#		}
		# Page 20h and 21h is only required if page 00h byte 195 bit 2 is high
#		set byteDec [lindex [my readRegister "00" 195] 0]
#		if {$byteDec & 1 == 0} {
#			eeprom removePage "20" "21"
#		}

#	}
	
	
	method isRegisterAccessSupported {} {
		# Check whether this interface has factory debug access
		# Use any function that accesses registers
		set err [catch {my getNominalRate} res]
		if {$err != 0} {
			cxLogDebug "isRegisterAccessSupported: NO"
			return False
		} else {
			return True
		}

	}
	method getType {} {
		return $Type
	}
	method getNumberOfChannels {} {
		return $NumberOfChannels
	}
	method getEepromPages {} {
		return [eeprom getPages]		
	}
	method getPortName {} {
		#set name [$Port getPortName]
		return $PortName
	}
	method getRegisterRange {page} {
#		return $NumberOfRegisters
		return [eeprom getRegisterRange $page]
	}
	method getAvailableMonitorItems {} {
		return [eeprom getAvailableMonitorItems]
	}
	method getMonitorData {item} {
		#puts "getMonitorData: $item"
		switch -- $item {
			RxPower	{
				return [my getRxPowers]
			}
			RxPowerAlarms {
				return [my getRxPowerAlarms]
			}
			TxPower	{
				return [my getTxPowers]
			}
			BIPS {
				return [my getBips]
			}
			LOS {
				return [my getLos]
			}
			FECcorr {
				return [my getFecCorrCounter]
			}
			FECuncorr {
				return [my getFecUncorrCounter]
			}
			FEClock {
				return [my getFecLockStatus]
			}
			default {
				return "$item: Not recognised"
			}
		}
		return
	}

	method getBips {} {
		set bipAddress [dict get $DeviceAddresses BIPaddr]
		if {$bipAddress == "NA"} {
			error "getbips: $Type not supported"
		}
		set bips [my readRegister BIP $bipAddress]
		set bipsHex [format "%02X" [lindex $bips 0]] 
		#set bips [lindex $bips 0] 
		return $bipsHex
	}
	method getFecCorrCounter {} {
		set fecAddress [dict get $DeviceAddresses FECcorr]
		if {$fecAddress == "NA"} {
			error "getFecCounter: $Type not supported"
		}
		set fecCounter [my readRegister FECcorr $fecAddress]
		set fecCounterHex [format "%02X" [lindex $fecCounter 0]] 
		#set fecCounter [lindex $fecCounter 0] 
		return $fecCounterHex
	}
	method getFecUncorrCounter {} {
		set fecAddress [dict get $DeviceAddresses FECuncorr]
		if {$fecAddress == "NA"} {
			error "getFecCounter: $Type not supported"
		}
		set fecCounter [my readRegister FECuncorr $fecAddress]
		set fecCounterHex [format "%02X" [lindex $fecCounter 0]] 
		#set fecCounter [lindex $fecCounter 0] 
		return $fecCounterHex
	}
	method getFecLockStatus {} {
		set fecAddress [dict get $DeviceAddresses FEClock]
		if {$fecAddress == "NA"} {
			error "getFecCounter: $Type not supported"
		}
		set fecCounter [my readRegister FEClock $fecAddress]
		set fecCounterHex [format "%02X" [lindex $fecCounter 0]] 
		#set fecCounter [lindex $fecCounter 0] 
		return $fecCounterHex
	}
	method getLos {} {
		set losReg [eeprom getLosRegister]
		set page [dict get $losReg "Page"]
		set addr [dict get $losReg "Address"]
		set los [my readRegister $page $addr]
		#set los [lindex [my readRegister $page $addr] 0]
		set losHex [format "%02X" [lindex $los 0]]
		return $losHex
	}
	method getRxPowerAlarms {} {
		set reg [eeprom getRxPowerAlarmRegisters]
		set page [dict get $reg "Page"]
		set addr [dict get $reg "Address"]
		set max [tcl::mathfunc::max {*}$addr]
		set min [tcl::mathfunc::min {*}$addr]

		set regRange [my readRegisterRange $page $min $max]
		
		set alarms ""
		foreach a $addr {
			#set alarm [lindex [my readRegister $page $a] 0]
			set alarm [dict get $regRange $a]
			set alarmHex [format "%02X" $alarm]
			append alarms $alarmHex " "
		}
		#set los [lindex [my readRegister $page $addr] 0]
		return $alarms
	}
	method GetPower_mW {page powerRegs} {
		set msbReg [dict get $powerRegs MSB]
		set lsbReg [dict get $powerRegs LSB]
		set msb [lindex [my readRegister $page $msbReg] 0]
		set lsb [lindex [my readRegister $page $lsbReg] 0]
		#set msb [expr 0x$msbHex]
		#set lsb [expr 0x$lsbHex]
		set power [expr [expr $msb << 8] + $lsb]
		# Units are 0.1uW, return as mW
		set power [expr $power * 0.0001]
		set power [format "%.2f" $power]
		return $power
	}
	method GetPowerFromRegRange_mW {regRange powerRegs} {
		set msbReg [dict get $powerRegs MSB]
		set lsbReg [dict get $powerRegs LSB]
		set msb [dict get $regRange $msbReg]
		set lsb [dict get $regRange $lsbReg]
		#set msb [expr 0x$msbHex]
		#set lsb [expr 0x$lsbHex]
		set power [expr [expr $msb << 8] + $lsb]
		# Units are 0.1uW, return as mW
		set power [expr $power * 0.0001]
		set power [format "%.2f" $power]
		return $power
	}
	method GetPowerRegisterRange {powerRegs txRx} {
		for {set i 1} {$i <= $NumberOfChannels} {incr i} {
			set regs [dict get $powerRegs "$txRx$i"]
			set msb [dict get $regs MSB]
			set lsb [dict get $regs LSB]
			lappend regList $msb $lsb
		}
		set max [tcl::mathfunc::max {*}$regList]
		set min [tcl::mathfunc::min {*}$regList]
		return "$min $max"
	}
	method getRxPowers {} {
		set powerList {}
		set powerRegs [eeprom getPowerRegisters]
		#set numberOfPowers [dict get $powerRegs "NumberOfPowers"]
		set page [dict get $powerRegs "Page"]
		set rxRange [my GetPowerRegisterRange $powerRegs Rx]
		# Now call readRegisterRange to get back the contents in a block
		set rxRegRange [my readRegisterRange $page [lindex $rxRange 0] [lindex $rxRange 1]]
		
		for {set i 1} {$i <= $NumberOfChannels} {incr i} {
			#set txPowerRegs [dict get $powerRegs "Tx$i"]
			set rxPowerRegs [dict get $powerRegs "Rx$i"]
			#set txPower [my GetPower_mW $page $txPowerRegs]
			#set rxPower [my GetPower_mW $page $rxPowerRegs]
			set rxPower [my GetPowerFromRegRange_mW $rxRegRange $rxPowerRegs]
			#set powers [list $rxPower $txPower]
			set powers [list $rxPower]
			lappend powerList $powers
		}
		return $powerList
	}
	method getTxPowers {} {
		set powerList {}
		set powerRegs [eeprom getPowerRegisters]
		#set numberOfPowers [dict get $powerRegs "NumberOfPowers"]
		set page [dict get $powerRegs "Page"]
		set txRange [my GetPowerRegisterRange $powerRegs Tx]
		# Now call readRegisterRange to get back the contents in a block
		set txRegRange [my readRegisterRange $page [lindex $txRange 0] [lindex $txRange 1]]

		for {set i 1} {$i <= $NumberOfChannels} {incr i} {
			#set txPowerRegs [dict get $powerRegs "Tx$i"]
			set txPowerRegs [dict get $powerRegs "Tx$i"]
			#set txPower [my GetPower_mW $page $txPowerRegs]
			#set txPower [my GetPower_mW $page $txPowerRegs]
			set txPower [my GetPowerFromRegRange_mW $txRegRange $txPowerRegs]
			#set powers [list $rxPower $txPower]
			set powers [list $txPower]
			lappend powerList $powers
		}
		return $powerList
	}
	method getPowers {} {
		cxLogDebug "[info level 0]"
		set powerList {}
		set powerRegs [eeprom getPowerRegisters]
		#set numberOfPowers [dict get $powerRegs "NumberOfPowers"]
		set page [dict get $powerRegs "Page"]
		set txRange [my GetPowerRegisterRange $powerRegs Tx]
		set rxRange [my GetPowerRegisterRange $powerRegs Rx]
		
		# Now call readRegisterRange to get back the contents in a block
		set txRegRange [my readRegisterRange $page [lindex $txRange 0] [lindex $txRange 1]]
		set rxRegRange [my readRegisterRange $page [lindex $rxRange 0] [lindex $rxRange 1]]
		
		for {set i 1} {$i <= $NumberOfChannels} {incr i} {
			set txPowerRegs [dict get $powerRegs "Tx$i"]
			set rxPowerRegs [dict get $powerRegs "Rx$i"]
			set txPower [my GetPowerFromRegRange_mW $txRegRange $txPowerRegs]
			set rxPower [my GetPowerFromRegRange_mW $rxRegRange $rxPowerRegs]
			set powers [list $rxPower $txPower]
			lappend powerList $powers
		}
		return $powerList
	}
	method getNominalRate {} {
		cxLogDebug "[info level 0]"
		set regs [eeprom getNominalRateRegisters]
		set regs1 [lindex $regs 0]
		set regs2 [lindex $regs 1]
		set rateDec [lindex [my readRegister [dict get $regs1 "Page"] [dict get $regs1 "Address"]] 0]
		#set rateDec [expr 0x$rateHex]
		if {$rateDec == 255} {
			set rateDec [lindex [my readRegister [dict get $regs2 "Page"] [dict get $regs2 "Address"]] 0]			
			#set rateDec [expr 0x$rateHex]
			set ret "[format "%0.2f" [expr $rateDec * 0.25]]G"
		} else {
			set ret "[format "%0.2f" [expr $rateDec * 0.1]]G"
		}
		cxLogDebug "getNominalRate: $ret"
		return $ret
	}
	method getEepromField {field} {
		cxLogDebug "[info level 0]"
		set ret "No decode"
		set val "--"
		set page "--"
		set addr "--"
		#if {[dict exists $eepromFields $field]} 
		if {[eeprom eepromFieldExists $field]} {
			set page [eeprom getEepromFieldInfo $field "Page"]
			set addr [eeprom getEepromFieldInfo $field "Address"]
			#set eepromField [dict get $eepromFields $field]
			#set addr [dict get $eepromField Address]
			set decval [lindex [my readRegister $page $addr] 0]
			set val [format "%02X" $decval]
			set addr [format "%02X" $addr]
			set ret [eeprom decodeFieldValue $field $decval]
			#set decodes [dict get $eepromField Decode]
			#if {[dict exists $decodes $val]} {
			#	set ret [dict get $decodes $val]
			#}
		}
		
		return "$ret ($page $addr:$val)"
	}
	#---------------------------------------------------------------------------
	# readRegister --
	# Arguments:
	#	device	Not sure what this is...
	#	addr	The module EEPROM address to be read
	#			For BIPS etc, this is hex; everything else is decimal
	# Results:
	#	Returns the contents of the address in hex
	#	Note: not sure what happens if there is no device inserted
	#	Note: There are two methods of reading registers. This one reads
	#	one at a time. See readRegisterStream for multiple registers
	method readRegister {page addrDec} {
		global eepromDict
		variable FpgaRegisters
		cxLogDebug "[info level 0]"
		
		set config "method put headers \{Content-Type application/json\}"
		set ret ""
		
		set deviceAddress [dict get $DeviceAddresses $page]
		if {$deviceAddress == "NA"} {
			error "readRegister: $Type not supported"
		}

		if {$page == "BIP" || $page == "FECcorr" || $page == "FECuncorr" || $page == "FEClock"} {
			#set addrHex $addr
			#set addrHex [format "%X" $addrDec]
			#set addrHex $addr
			set TypeForRest "FPGA"
			set regName $page
			set deviceAddress 0
			set regVal [FpgaRegisters registerGet2 $page $deviceAddress $addrDec]			
		} else {
			#set addrHex [format "%X" $addr]
			
			set TypeForRest [dict get $moduleDict $Type typeForRest]
			set regNames [eeprom getModuleDetails "RegNames" $page]
			set regName ""
			if {[dict exists $regNames $addrDec]} {
				set regName [dict get $regNames $addrDec]
			}

			# For QSFP28, there are potentially multiple pages so we need to write
			# to the page select register. This is done later if required.
			# If the page select does not need to change, then use FF
			#set page [my WritePageSelect $page]
			set page [my GetPageSelectValue $page]
			set regVal [ModuleRegisters registerGet2 $page $deviceAddress $addrDec]			
		}
		
		set ret [lappend ret $regVal $regName]
		cxLogDebug "readRegister return: $ret"
		return $ret
	}
	
	method readRegisterRange {page minDec maxDec} {
		set deviceAddress [dict get $DeviceAddresses $page]
		if {$deviceAddress == "NA"} {
			error "readRegisterRange: $Type not supported"
		}

		if {$page == "BIP" || $page == "FECcorr" || $page == "FECuncorr" || $page == "FEClock"} {
			error "readRegisterRange not supported for $page"
		} else {
			
			#set minHex [format "%X" $minDec]
			#set maxHex [format "%X" $maxDec]
			set TypeForRest [dict get $moduleDict $Type typeForRest]
			#set regNames [eeprom getModuleDetails "RegNames" $page]
			#set regName ""
			#if {[dict exists $regNames $addr]} {
			#	set regName [dict get $regNames $addr]
			#}
			#set page [my WritePageSelect $page]
			set page [my GetPageSelectValue $page]
			set valList [ModuleRegisters registerGetRange $page $deviceAddress $minDec $maxDec]			
		}
	
		return $valList
	}

	method openRegisterReader {page startAddrDec endAddrDec} {
		cxLogDebug "$page $startAddrDec $endAddrDec"
		
		if {$page == "BIP" || $page == "FECcorr" || $page == "FECuncorr" || $page == "FEClock"} {
			error "openRegisterReader: $page not supported"
		} else {
			set deviceAddress [dict get $DeviceAddresses $page]
			if {$deviceAddress == "NA"} {
				error "readRegister: $Type not supported"
			}

			#set startAddrHex [format "%X" $startAddrDec]
			#set endAddrHex [format "%X" $endAddrDec]
			set RegisterReaderPage $page
			#set page [my WritePageSelect $page]
			set page [my GetPageSelectValue $page]
			set RegisterReaderFileHandle [ModuleRegisters registerReaderOpen $page $deviceAddress $startAddrDec $endAddrDec]
		}
	}
	method closeRegisterReader {} {
		cxLogDebug ""
		set RegisterReaderFileHandle ""
		ModuleRegisters registerReaderClose
	}
	# Read a line from the reader
	# Each line is in the form <addr> <value>
	# For each <addr>, get the decode and return <addr> <value> <decode>
	method getRegisterReaderLine {} {
		cxLogDebug ""
		set retval ""
		if {$RegisterReaderFileHandle == ""} {
			error "getRegisterReaderLine: no reader open"
		} else {
			set ret [ModuleRegisters registerReaderGet]
			set addr [lindex $ret 0]
			set val [lindex $ret 1]
			set regNames [eeprom getModuleDetails "RegNames" $RegisterReaderPage]
			set regName ""
			if {[dict exists $regNames $addr]} {
				set regName [dict get $regNames $addr]
			}
			lappend retval $addr $val $regName
		}
		cxLogDebug "$retval"
		return $retval
	}
	
	method readGpio {} {
		cxLogDebug "[info level 0]"
		set device [dict get $gpioSettings "device"]
		set bank [dict get $gpioSettings "bank"]
		set pin [dict get $gpioSettings "pin"]
		set ret [GpioRegisters registerGet2 "100" $device $bank "getPinValue" $pin]
		return $ret
	}
	
	method writeRegister {page addrDec dataDec} {
		cxLogDebug "[info level 0]"

		set deviceAddress [dict get $DeviceAddresses $page]
		if {$deviceAddress == "NA"} {
			error "writeRegister: $Type not supported"
		}

		#set addrHex [format "%X" $addrDec]
		#set page [my WritePageSelect $page]
		set page [my GetPageSelectValue $page]
		set regVal [ModuleRegisters registerSet2 $page $deviceAddress $addrDec $dataDec]
		set retval [dict get $regVal "value"]
		cxLogDebug "writeRegister Return: $retval"
		return $retval
	}
	
	method GetPageSelectValue {page} {
		cxLogDebug "[info level 0]"
		set retval $page

		if {$Type == "QSFP28" || $Type == "QSFPPLUS"} {
			if {$CurrentPageSelect != $page} {
				set CurrentPageSelect $page
				set retval $page
			} else {
				set retval "100"				
			}
		} else {
			set CurrentPageSelect $page
			set retval $page
		}

		cxLogDebug "getPageSelectValue Return: $retval"
		return $retval
		
	}
		
	method isPageSupported {page} {
		cxLogDebug "[info level 0]"
		set retval False
		if {$Type == "QSFP28" || $Type == "QSFPPLUS"} {
			set pageDec [expr 0x$page]
			set regVal [my writeRegister "00" 127 $pageDec]
			#set valueHex [format %X $regVal]
			if {$regVal == $pageDec} {
				set retval True
			}
		} else {
			set retval True
		}
		cxLogDebug "isPageSupported Return: $retval"
		return $retval
	}

}
