################################################################################
# CxInstrumentOpticalModuleEepromClass.tcl
# ------------------------------------------------------------------------------
# This script is provided as a utility supporting Paragon-neo and Paragon-100G.
# It is NOT intended as example code for controlling these instruments.
# No part of this script should be used independently of this script without
# the explicit written permission of Calnex Solutions Ltd.
################################################################################
source calnexRest.tcl

oo::class create CxInstrumentOpticalModuleEeprom {
	variable eepromDict
	variable eepromFields
	variable Pages
	variable Type

	constructor {type} {
#			"Port 1" {"A0" 1 "A2" 0} 
#			"Port 2" {"A0" 4 "A2" 3} 
		# SFP28 defined by SFF-8472
		dict set eepromDict "SFP28" { \
			"Pages" {"A0" "A2"} \
			"RegisterRange" {"A0" {first 0 last 95} "A2" {first 0 last 127}} \
			"Power" {Page "A2" Tx1 {MSB 102 LSB 103} Rx1 {MSB 104 LSB 105}} \
			"RxPowerAlarms" {Page "A2" Address 113} \
			"AvailableMonitorItems" {"RxPower" "TxPower" "RxPowerAlarms" "BIPS" "FECcorr" "FECuncorr" "FEClock"}
			"RegNames" { \
				"A0" {0 "Type" 1 "Ext Id" 3 "10G/InfiniBand Compliance Codes" \
						6 "Ethernet Compliance Codes" 7 "Fibre Channel Link Length" \
						11 "Encoding" 12 "BR Nominal" 13 "Rate Identifier" \
						20 "Start of Vendor Name" 35 "End of Vendor Name" \
						36 "Ext. Compliance" 37 "Start of Vendor OUI" 39 "End of Vendor OUI" \
						40 "Start of Vendor PN" 55 "End of Vendor PN" 56 "Start of Vendor Rev" 59 "End of Vendor Rev" \
						63 "CC_BASE" 64 "Options #1" 65 "Options #2" 66 "BR Max" 67 "BR Min" \
						68 "Start of Vendor SN" 83 "End of Vendor SN"  84 "Start of Date Code" 91 "End of Date Code" \
						92 "Diagnostic Monitoring Type" 93 "Enhanced Options"
						94 "SFF-8472 Compliance" 95 "CC_EXT"} \
				"A2" {00 "Start of Alarm and Warning Thresholds" 55 "End of Alarm and Warning Thresholds" \
						56 "Start of Calibration Constants" 91 "End of Calibration Constants" \
						92 "Unallocated" 93 "Unallocated" 94 "Unallocated" 
						95 "CC_DMI" 96 "Temp MSB" 97 "Temp LSB" 98 "Vcc MSB" 99 "Vcc LSB" 100 "Tx Bias MSB" 101 "Tx Bias LSB" \
						102 "Tx Power MSB" 103 "Tx Power LSB" 104 "Rx Power MSB" 105 "Rx Power LSB" \
						106 "Laser Temp/Wavelength MSB" 107 "Laser Temp/Wavelength MSB" \
						108 "TEC Current MSB" 109 "TEC Current LSB" 110 "Optional Status/Control" \
						112 "Alarms #1" 113 "Alarms #2" 114 "Alarms #3" 115 "Alarms #4" 116 "Alarms #5" 117 "Alarms #6" \
						119 "CDR Lock"}}}
		# SFP+ is also defined by SFF-8472 so copy
		#dict set eepromDict "SFPPLUS" [dict get $eepromDict "SFP28"]
		# ... but change the available monitor items
		dict set eepromDict "SFPPLUS" [dict replace [dict get $eepromDict "SFP28"] "AvailableMonitorItems" {"RxPower" "TxPower" "RxPowerAlarms" "BIPS"}]
		# SFP1G is also defined by SFF-8472 so copy
		dict set eepromDict "SFP1G" [dict replace [dict get $eepromDict "SFP28"] "AvailableMonitorItems" {"RxPower" "TxPower" "RxPowerAlarms" "BIPS"}]

		# QSFP28 defined by SFF-8636
#			"Port 1" {"00" 0} 
#			"Port 2" {"00" 1} 
		dict set eepromDict "QSFP28" { \
			"Pages" {"00" "01" "02" "03" "20" "21"} \
			"RegisterRange" { "00" {first 0 last 255} "01" {first 127 last 255} "02" {first 127 last 255} "03" {first 127 last 255} "20" {first 127 last 255} "21" {first 127 last 255}} \
			"Power" {Page "00" \
						Tx1 {MSB 50 LSB 51} Tx2 {MSB 52 LSB 53} Tx3 {MSB 54 LSB 55} Tx4 {MSB 56 LSB 57}\
						Rx1 {MSB 34 LSB 35} Rx2 {MSB 36 LSB 37} Rx3 {MSB 38 LSB 39} Rx4 {MSB 40 LSB 41}} \
			"LOS" {Page "00" Address 3} \
			"RxPowerAlarms" {Page "00" Address {9 10}} \
			"AvailableMonitorItems" {"RxPower" "TxPower" "RxPowerAlarms" "LOS" "BIPS" "FECcorr" "FECuncorr" "FEClock"} \
			"RegNames" { 
				"00" {0 "Identifier" 1 "Revision Compliance" 3 "LOS" 4 "Tx Faults" 5 "CDR LOL" 6 "Temp Alarms" 7 "Vcc Alarms" 9 "Rx Power Alarms #1" 10 "Rx Power Alarms #2" \
						11 "Tx Bias Alarms #1" 12 "Tx Bias Alarms #2" 13 "Tx Power Alarms #1" 14 "Tx Power Alarms #2" \
						22 "Temperature MSB" 23 "Temperature LSB" 26 "Supply Voltage MSB" 27 "Supply Voltage LSB" \
						34 "Rx1 Power MSB" 35 "Rx1 Power LSB" 36 "Rx2 Power MSB" 37 "Rx2 Power LSB" 38 "Rx3 Power MSB" 39 "Rx3 Power LSB" 40 "Rx4 Power MSB" 41 "Rx4 Power LSB" \
						50 "Tx1 Power MSB" 51 "Tx1 Power LSB" 52 "Tx2 Power MSB" 53 "Tx2 Power LSB" 54 "Tx3 Power MSB" 55 "Tx3 Power LSB" 56 "Tx4 Power MSB" 57 "Tx4 Power LSB" \
						74 "Start of Vendor Specific" 81 "End of Vendor Specific" \
						86 "Tx Disable" 87 "Rx Rate Select" 88 "Tx Rate Select" 93 "High Power Class Enable" 98 "CDR Control" \
						103 "Temp Alarms" 104 "Vcc Alarms" 108 "Propagation Delay MSB" 109 "Propagation Delay LSB" 110 "Advanced Low Power Mode" 127 "Page Select" \
						128 "Identifier" 129 "Ext. Identifier" 130 "Connector Type" 131 "10G/40G Ethernet Compliance Codes" 134 "Gigabit Ethernet Compliance Codes" 139 "Encoding" \
						140 "BR Nominal (100MBd)" 141 "External Rate Select Compliance" 142 "Length (SMF)" 147 "Device Technology" 148 "Start of Vendor Info" 164 "Extended Module" 185 "End of Vendor Info" \
						186 "Wavelength or Copper Atten" 187 "Wavelength or Copper Atten" 188 "Wavelength Tol or Copper Atten" 189 "Wavelength Tol or Copper Atten" 191 "CC_BASE" \
						192 "Ext. Compliance" 193 "Tx IP Eq / Rx OP Settings" 194 "Tx/Rx CDR and squelch options" 195 "Pages/RATE_SELECT/other options" \
						196 "Start of Vendor SN" 211 "End of Vendor SN" 212 "Start of Date Code" 219 "End of Date Code" \
						220 "Diagnostic Monitoring Type" 221 "Enhanced Options" 222 "BR Nominal (250MBd)"} \
				"01" {128 "Start of module identification info" 255 "End of module identification info"} \
				"02" {128 "Start of user writeable memory" 255 "End of user writeable memory"} \
				"03" {128 "Start of optional channel controls" 255 "End of optional channel controls"} \
				"20" {128 "Start of interrupt mask bits for monitored parameters" 255 "End of interrupt mask bits for monitored parameters"} \
				"21" {128 "Start of threshold registers for monitored parameters" 255 "End of threshold registers for monitored parameters"}} \
		}
		# QSFP+ is defined by SFF-8636 so copy.
		# Note that there is an SFF-8436 that was specifically for QSFP+
		# but this appears to have been superceded by 8636 (see SFF-8024 for details)
		#dict set eepromDict "QSFPPLUS" [dict get $eepromDict "QSFP28"]
		dict set eepromDict "QSFPPLUS" [dict replace [dict get $eepromDict "QSFP28"] "AvailableMonitorItems" {"RxPower" "TxPower" "RxPowerAlarms" "BIPS"}]
#			"Port 1" {"00" 0} 
#			"Port 2" {"00" 1} 
		#dict set eepromDict "QSFPPLUS" { \
		#	"Pages" {"00"} \
		#	"RegisterRange" { "00" {first 0 last 222}} \
		#	"Power" {Page "00" \
		#				Tx1 {MSB 50 LSB 51} Tx2 {MSB 52 LSB 53} Tx3 {MSB 54 LSB 55} Tx4 {MSB 56 LSB 57}\
		#				Rx1 {MSB 34 LSB 35} Rx2 {MSB 36 LSB 37} Rx3 {MSB 38 LSB 39} Rx4 {MSB 40 LSB 41}} \
		#	"RegNames" { \
		#		"00" {0 "Identifier" 3 "LOS" 4 "Tx Faults" 6 "Temp Alarms" 7 "Vcc Alarms" 9 "Rx Power Alarms #1" 10 "Rx Power Alarms #2" \
		#				11 "Tx Bias Alarms #1" 12 "Tx Bias Alarms #2" 13 "Tx Power Alarms #1" 14 "Tx Power Alarms #2" \
		#				26 "Supply Voltage MSB" 27 "Supply Voltage LSB" \
		#				34 "Rx1 Power MSB" 35 "Rx1 Power LSB" 36 "Rx2 Power MSB" 37 "Rx2 Power LSB" 38 "Rx3 Power MSB" 39 "Rx3 Power LSB" 40 "Rx4 Power MSB" 41 "Rx4 Power LSB" \
		#				86 "Tx Disable" 87 "Rx Rate Select" 88 "Tx Rate Select" \
		#				127 "Page Select" \
		#				128 "Identifier" 129 "Ext. Identifier" 131 "10G/40G Ethernet Compliance Codes" 134 "Gigabit Ethernet Compliance Codes" 139 "Encoding" \
		#				140 "BR Nominal (100MBd)" 141 "External Rate Select Compliance" 220 "Diagnostic Monitoring Type"}} \
		#}
		dict set eepromDict "CXP" 			{"Pages" {} "AvailableMonitorItems" NONE}
		dict set eepromDict "CFP2" 			{"Pages" {} "AvailableMonitorItems" NONE}
		dict set eepromDict "CFP4" 			{"Pages" {} "AvailableMonitorItems" {"FECcorr" "FECuncorr" "FEClock"}}
		dict set eepromDict "RJ45_1G"		{"Pages" {} "AvailableMonitorItems" NONE}
		dict set eepromDict "RJ45_LOWRATE"	{"Pages" {} "AvailableMonitorItems" NONE}
		#dict set eepromDict "SFP1G"		{"Pages" {} "AvailableMonitorItems" NONE}
		dict set eepromDict "SFP100M"		{"Pages" {} "AvailableMonitorItems" NONE}
		dict set eepromDict "NONE"			{"Pages" {} "AvailableMonitorItems" NONE}

		#dict set eepromDict "QSFPPLUS" { \
		#	"RegisterRange" {first 128 last 222}
		#	"Port 1" 0 \
		#	"Port 2" 1 \
		#	"RegNames" { \
		#		128 "Type" 131 "10G/40G Enthernet Compliance Code" 139 "Encoding" \
		#		140 "BR Nominal (100MBd)" 141 "External Rate Select Compliance" 192 "Ext. Compliance" 222 "BR Nominal (250MBd)"} \
		#}

		#dict set eepromFields "SFPPLUS" {\
		#	Type 			{Address 0 Decode {00 Unspecified 03 "SFP/SFP+/SFP28" 06 XFP 0C QSFP 0D "QSFP+" 0E CXP 11 QSFP28 12 CXP2}} \
		#	Encoding 		{Address 6 Decode {00 Unspecified 01 "8B/10B" 02 "4B/5B" 03 "NRZ" 04 Manchester 05 SONET 06 "64B/66B" 07 "256B/257B" 08 "PAM4"}} \
		#	ExtCompliance 	{Address 36 Decode {00 Unspecified 01 AOC1 02 "100GBASE-SR4/25GBASE-SR" 03 "100GBASE-LR4/25GBASE-LR" 04 "100GBASE-ER4/25GBASE-ER" 05 "100GBASE-SR10" \
		#										06 "100G CWDM4" 07 "100G PSM4" 08 "ACC1" 09 Obsolete 0A Reserved 0B "100GBASE-CR4/25GBASE-CR CA-L" 0C "25GBASE-CR CA-S" 0C "25GBASE-CR CA-N" \
		#										10 "40GBASE-ER4" 11 "4x10GBASE-SR" 12 "40G PSM4" 16 "10GBASE-T (SFI)" 17 "100G CLR4" 18 AOC2 19 ACC2 1A "100GE-DWDM2"}} \
		#	BrNominal 		{Address 12} \
		#} 
		dict set eepromFields "SFP28" 	{\
			Type 			{Page "A0" Address 0 Decode {00 Unspecified 03 "SFP/SFP+/SFP28" 06 XFP 0C QSFP 0D "QSFP+" 0E CXP 11 QSFP28 12 CXP2}}
			Encoding 		{Page "A0" Address 11 Decode {00 Unspecified 01 "8B/10B" 02 "4B/5B" 03 "NRZ" 04 Manchester 05 SONET 06 "64B/66B" 07 "256B/257B" 08 "PAM4"}} 
 			ExtCompliance 	{Page "A0" Address 36 Decode {00 Unspecified 01 AOC1 02 "100GBASE-SR4/25GBASE-SR" 03 "100GBASE-LR4/25GBASE-LR" 04 "100GBASE-ER4/25GBASE-ER" 05 "100GBASE-SR10" \
												06 "100G CWDM4" 07 "100G PSM4" 08 "ACC1" 09 Obsolete 0A Reserved 0B "100GBASE-CR4/25GBASE-CR CA-L" 0C "25GBASE-CR CA-S" 0C "25GBASE-CR CA-N" \
												10 "40GBASE-ER4" 11 "4x10GBASE-SR" 12 "40G PSM4" 16 "10GBASE-T (SFI)" 17 "100G CLR4" 18 AOC2 19 ACC2 1A "100GE-DWDM2"}}
			BrNominal1 		{Page "A0" Address 12}\
			BrNominal2 		{Page "A0" Address 66}\
		}
		dict set eepromFields "SFPPLUS" [dict get $eepromFields "SFP28"]
		dict set eepromFields "SFP1G" [dict get $eepromFields "SFP28"]

		#dict set eepromFields "QSFPPLUS" 	{\
		#	Type 			{Address  128 Decode {00 Unspecified 03 "SFP/SFP+/SFP28" 06 XFP 0C QSFP 0D "QSFP+" 0E CXP 11 QSFP28 12 CXP2}} 
		#	Encoding 		{Address 139 Decode {00 Unspecified 01 "8B/10B" 02 "4B/5B" 03 "NRZ" 04 SONET 05 "64B/66B" 06 Manchester 07 "256B/257B" 08 "PAM4"}} 
 		#	ExtCompliance 	{Address 192 Decode {00 Unspecified 01 AOC1 02 "100GBASE-SR4/25GBASE-SR" 03 "100GBASE-LR4/25GBASE-LR" 04 "100GBASE-ER4/25GBASE-ER" 05 "100GBASE-SR10" \
		#										06 "100G CWDM4" 07 "100G PSM4" 08 "ACC1" 09 Obsolete 0A Reserved 0B "100GBASE-CR4/25GBASE-CR CA-L" 0C "25GBASE-CR CA-S" 0C "25GBASE-CR CA-N" \
		#										10 "40GBASE-ER4" 11 "4x10GBASE-SR" 12 "40G PSM4" 16 "10GBASE-T (SFI)" 17 "100G CLR4" 18 AOC2 19 ACC2 1A "100GE-DWDM2"}}
		#	BrNominal 		{Address 140}\
		#} 
		dict set eepromFields "QSFP28" 	{\
			Type 			{Page "00" Address 128 Decode {00 Unspecified 03 "SFP/SFP+/SFP28" 06 XFP 0C QSFP 0D "QSFP+" 0E CXP 11 QSFP28 12 CXP2}}
			Encoding 		{Page "00" Address 139 Decode {00 Unspecified 01 "8B/10B" 02 "4B/5B" 03 "NRZ" 04 SONET 05 "64B/66B" 06 Manchester 07 "256B/257B" 08 "PAM4"}} 
 			ExtCompliance 	{Page "00" Address 192 Decode {00 Unspecified 01 AOC1 02 "100GBASE-SR4/25GBASE-SR" 03 "100GBASE-LR4/25GBASE-LR" 04 "100GBASE-ER4/25GBASE-ER" 05 "100GBASE-SR10" \
												06 "100G CWDM4" 07 "100G PSM4" 08 "ACC1" 09 Obsolete 0A Reserved 0B "100GBASE-CR4/25GBASE-CR CA-L" 0C "25GBASE-CR CA-S" 0C "25GBASE-CR CA-N" \
												10 "40GBASE-ER4" 11 "4x10GBASE-SR" 12 "40G PSM4" 16 "10GBASE-T (SFI)" 17 "100G CLR4" 18 AOC2 19 ACC2 1A "100GE-DWDM2"}}
			BrNominal1 		{Page "00" Address 140}\
			BrNominal2		{Page "00" Address 222}\
		}
		dict set eepromFields "QSFPPLUS" [dict get $eepromFields "QSFP28"]

		# HACK HACK HACK - there is no BrNominal2 register in QSFPPLUS - leave it here for the moment - it is expected to exist in the rest of the code		
#		dict set eepromFields "QSFPPLUS" 	{\
#			Type 			{Page "00" Address 128 Decode {00 Unspecified 03 "SFP/SFP+/SFP28" 06 XFP 0C QSFP 0D "QSFP+" 0E CXP 11 QSFP28 12 CXP2}}
#			Encoding 		{Page "00" Address 139 Decode {00 Unspecified 01 "8B/10B" 02 "4B/5B" 03 "NRZ" 04 SONET 05 "64B/66B" 06 Manchester 07 "256B/257B" 08 "PAM4"}} 
#			BrNominal1 		{Page "00" Address 140}\
#			BrNominal2		{Page "00" Address 222}\
#		}

		#puts "Creating optical module eeprom: $type"

		set Type [my MapTypes $type]
		set Pages [dict get $eepromDict $Type "Pages"]
	}
	destructor {}
	
	method MapTypes {type} {
#		if {$type == "SFPPLUS"} {
#			return "SFP28"
#		}
#		if {$type == "QSFPPLUS"} {
#			return "QSFP28"
#		}
		return $type
	}
	method removePage {pages} {
		if {$Type == "QSFP28" || $Type == "QSFPPLUS"} {
			foreach p $pages {
				set i [lsearch -exact $Pages $p]
				if {i >= 0} {
					set Pages [lreplace $Pages $i $i]
				}				
			}
		}
	}
	
	method getPages {} {
		return $Pages		
	}
	method getModuleDetails {args} {
		return [dict get $eepromDict $Type {*}$args]
	}
	method getEepromFieldInfo {args} {
		return [dict get $eepromFields $Type {*}$args]
	}
	method eepromFieldExists {field} {
		set fields [dict get $eepromFields $Type]
		return [dict exists $fields $field]
	}
	method getRegisterRange {page} {
		return [dict get $eepromDict $Type "RegisterRange" $page]
	}
	method getAvailableMonitorItems {} {
		return [dict get $eepromDict $Type "AvailableMonitorItems"]
	}
	method decodeFieldValue {field value} {
		set ret "No decode"
		#set val "--"
		set val [format "%02X" $value]
		if {$field == "BrNominal"} {
			if {$value >= 255} {
				set ret ">25.0G"
			} else {
				set ret "[expr $value * 0.1]G"
			}
		} else {
			set decodes [my getEepromFieldInfo $field Decode]
			#set decodes [dict get $eepromField Decode]
			if {[dict exists $decodes $val]} {
				set ret [dict get $decodes $val]
			}
		}
		return $ret
	}
	method getNominalRateRegisters {} {
		set reg1 [dict get $eepromFields $Type "BrNominal1"]		
		set reg2 [dict get $eepromFields $Type "BrNominal2"]		
		return [list $reg1 $reg2]
	}
	method getPowerRegisters {} {
		return [dict get $eepromDict $Type "Power"]
	}
	method getLosRegister {} {
		return [dict get $eepromDict $Type "LOS"]
	}
	method getRxPowerAlarmRegisters {} {
		return [dict get $eepromDict $Type "RxPowerAlarms"]
	}
}
