################################################################################
# CxRegisters.tcl
# ------------------------------------------------------------------------------
# This script is provided as a utility supporting Paragon-neo and Paragon-100G.
# It is NOT intended as example code for controlling these instruments.
# No part of this script should be used independently of this script without
# the explicit written permission of Calnex Solutions Ltd.
################################################################################
#
# Read and write FPGA registers in Paraon-100G / Paragon-neo
#
# Notes:
#	Use of these commands may result in an unstable or damaged instrument
# 	Do not use these commands unless requested to do so by Calnex
#
source calnexRest.tcl

oo::class create CxRegister {
	variable AssemblyId
	variable DeviceId
	variable ReaderFileHandle
	variable RegisterReadWriteLogFile
	
	constructor {assyId devId} {
		cxLogDebug "[info level 0]"
		set AssemblyId $assyId
		set DeviceId $devId
		set date [clock format [clock seconds] -format %Y-%m-%d_%H-%M-%S]
		set RegisterReadWriteLogFile "CxDiagLogging/RegisterReadWrite_${assyId}_${devId}_${date}.txt"
	}
	
	destructor {}
	
	# The address to be written is decimal
	# The value to be wrriten is decimal
	method registerSet2 {page deviceNumber addrDec valueDec} {
		cxLogDebug "[info level 0]"
		set ipAddr [calnexGetIpAddr]
		# Page is not required for FPGA or DIGIO
		if {$DeviceId == "FPGA" || $DeviceId == "DIGIO"} {
			set pageDec 256
		} else {
			set pageDec [expr 0x$page]
		}
		#set addrDec [expr 0x$addrHex]
		set io [open "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c writeRegister -j $pageDec -s $addrDec -v $valueDec -l $RegisterReadWriteLogFile"]
		set len [gets $io line]
		cxLogDebug "Written: $line (len=$len)"
		close $io
		return $line
	}
	
	# The address to be read is decimal
	# The value returned is decimal
	method registerGet2 {page deviceNumber addrDec {commandId "readRegister"} {parameter2 ""}} {
		cxLogDebug "[info level 0]"
		set ipAddr [calnexGetIpAddr]

		# Page is not required for FPGA or DIGIO
		if {$DeviceId == "FPGA" || $DeviceId == "DIGIO"} {
			set pageDec 256
		} else {
			set pageDec [expr 0x$page]
		}

		#set addrDec [expr 0x$addrHex]
		if {$parameter2 == ""} {
			cxLogDebug "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c $commandId -j $pageDec -s $addrDec -l $RegisterReadWriteLogFile"
			set io [open "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c $commandId -j $pageDec -s $addrDec -l $RegisterReadWriteLogFile"]
		} else {
			cxLogDebug "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c $commandId -j $pageDec -s $addrDec -p $parameter2 -l $RegisterReadWriteLogFile"
			set io [open "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c $commandId -j $pageDec -s $addrDec -p $parameter2 -l $RegisterReadWriteLogFile"]			
		}
		set len [gets $io line]
		cxLogDebug "Read: $line (len=$len)"
		#set parts [split $line ":"]
		#set retval [lindex $parts 1]
		close $io
		set regVal [dict get $line "value"]
		#cxLogDebug "$regVal"
		return $regVal
	}
	method registerGetRange {page deviceNumber startAddrDec endAddrDec} {
		cxLogDebug "[info level 0]"
		set ipAddr [calnexGetIpAddr]
		# Page is not required for FPGA or DIGIO
		if {$DeviceId == "FPGA" || $DeviceId == "DIGIO"} {
			set pageDec 256
		} else {
			set pageDec [expr 0x$page]
		}
		#set startAddrDec [expr 0x$startAddrHex]
		#set endAddrDec [expr 0x$endAddrHex]
		cxLogDebug "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c readRegister -j $pageDec -s $startAddrDec -e $endAddrDec -l $RegisterReadWriteLogFile"
		set io [open "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c readRegister -j $pageDec -s $startAddrDec -e $endAddrDec -l $RegisterReadWriteLogFile"]

		for {set a $startAddrDec} {$a < [expr $endAddrDec+1]} {incr a} {
			set len [gets $io line]
			cxLogDebug "Read: $line (len=$len)"
			#set parts [split $line ":"]
			#set val [lindex $parts 1]
			#cxLogDebug "$line"
			
			set address [dict get $line "address"]
			set value [dict get $line "value"]
			#lappend retval $line
			lappend retval $address $value
		}
		close $io
		
		return $retval
	}
	
	method registerReaderOpen {page deviceNumber startAddrDec endAddrDec} {
		cxLogDebug "[info level 0]"
		set ipAddr [calnexGetIpAddr]
		set pageDec [expr 0x$page]
		#set startAddrDec [expr 0x$startAddrHex]
		#set endAddrDec [expr 0x$startAddrHex]

		set ReaderFileHandle [open "|dist/registerReadWrite.exe -i $ipAddr -a $AssemblyId -d $DeviceId -n $deviceNumber -c readRegister -j $pageDec -s $startAddrDec -e $endAddrDec -l $RegisterReadWriteLogFile"]
		
		return $ReaderFileHandle
	}
	
	method registerReaderClose {} {
		cxLogDebug "[info level 0]"
		close $ReaderFileHandle
	}
	
	method registerReaderGet {} {
		cxLogDebug "[info level 0]"
		# Should check for file handle being open
		set len [gets $ReaderFileHandle line]
		cxLogDebug "Read: $line (len=$len)"
		
		set address [dict get $line "address"]
		set value [dict get $line "value"]
		lappend retval $address $value

		return $retval
	}
}

